import time
from abc import ABC, abstractmethod
from collections import defaultdict

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from sklearn.manifold import TSNE
from torch.utils.data import Dataset

matplotlib.use("Agg")
plt.switch_backend("Agg")

import logging

from nebula.config.config import TRAINING_LOGGER
from nebula.core.utils.deterministic import enable_deterministic

logging_training = logging.getLogger(TRAINING_LOGGER)


class NebulaDataset(Dataset, ABC):
    """
    Abstract class for a partitioned dataset.

    Classes inheriting from this class need to implement specific methods
    for loading and partitioning the dataset.
    """

    def __init__(
        self,
        num_classes=10,
        partition_id=0,
        partitions_number=1,
        batch_size=32,
        num_workers=4,
        iid=True,
        partition="dirichlet",
        partition_parameter=0.5,
        seed=42,
        config=None,
    ):
        super().__init__()

        if partition_id < 0 or partition_id >= partitions_number:
            raise ValueError(f"partition_id {partition_id} is out of range for partitions_number {partitions_number}")

        self.num_classes = num_classes
        self.partition_id = partition_id
        self.partitions_number = partitions_number
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.iid = iid
        self.partition = partition
        self.partition_parameter = partition_parameter
        self.seed = seed
        self.config = config

        self.train_set = None
        self.train_indices_map = None
        self.test_set = None
        self.test_indices_map = None

        # Classes of the participants to be sure that the same classes are used in training and testing
        self.class_distribution = None

        enable_deterministic(config)

        if self.partition_id == 0:
            self.initialize_dataset()
        else:
            max_tries = 10
            for i in range(max_tries):
                try:
                    self.initialize_dataset()
                    break
                except Exception as e:
                    logging_training.info(f"Error loading dataset: {e}. Retrying {i + 1}/{max_tries} in 5 seconds...")
                    time.sleep(5)

    @abstractmethod
    def initialize_dataset(self):
        """
        Initialize the dataset. This should load or create the dataset.
        """
        pass

    @abstractmethod
    def generate_non_iid_map(self, dataset, partition="dirichlet", plot=False):
        """
        Create a non-iid map of the dataset.
        """
        pass

    @abstractmethod
    def generate_iid_map(self, dataset, plot=False):
        """
        Create an iid map of the dataset.
        """
        pass

    def get_train_labels(self):
        """
        Get the labels of the training set based on the indices map.
        """
        if self.train_indices_map is None:
            return None
        return [self.train_set.targets[idx] for idx in self.train_indices_map]

    def get_test_labels(self):
        """
        Get the labels of the test set based on the indices map.
        """
        if self.test_indices_map is None:
            return None
        return [self.test_set.targets[idx] for idx in self.test_indices_map]

    def get_local_test_labels(self):
        """
        Get the labels of the local test set based on the indices map.
        """
        if self.local_test_indices_map is None:
            return None
        return [self.test_set.targets[idx] for idx in self.local_test_indices_map]

    def plot_data_distribution(self, dataset, partitions_map):
        """
        Plot the data distribution of the dataset.

        Plot the data distribution of the dataset according to the partitions map provided.

        Args:
            dataset: The dataset to plot (torch.utils.data.Dataset).
            partitions_map: The map of the dataset partitions.
        """
        # Plot the data distribution of the dataset, one graph per partition
        sns.set()
        sns.set_style("whitegrid", {"axes.grid": False})
        sns.set_context("paper", font_scale=1.5)
        sns.set_palette("Set2")

        for i in range(self.partitions_number):
            indices = partitions_map[i]
            class_counts = [0] * self.num_classes
            for idx in indices:
                label = dataset.targets[idx]
                class_counts[label] += 1
            logging_training.info(f"Participant {i + 1} class distribution: {class_counts}")
            plt.figure()
            plt.bar(range(self.num_classes), class_counts)
            plt.xlabel("Class")
            plt.ylabel("Number of samples")
            plt.xticks(range(self.num_classes))
            if self.iid:
                plt.title(f"Participant {i + 1} class distribution (IID)")
            else:
                plt.title(
                    f"Participant {i + 1} class distribution (Non-IID - {self.partition}) - {self.partition_parameter}"
                )
            plt.tight_layout()
            path_to_save = f"{self.config.participant['tracking_args']['log_dir']}/{self.config.participant['scenario_args']['name']}/participant_{i}_class_distribution_{'iid' if self.iid else 'non_iid'}{'_' + self.partition if not self.iid else ''}.png"
            plt.savefig(path_to_save, dpi=300, bbox_inches="tight")
            plt.close()

        plt.figure()
        max_point_size = 500
        min_point_size = 0

        for i in range(self.partitions_number):
            class_counts = [0] * self.num_classes
            indices = partitions_map[i]
            for idx in indices:
                label = dataset.targets[idx]
                class_counts[label] += 1

            # Normalize the point sizes for this partition
            max_samples_partition = max(class_counts)
            sizes = [
                (size / max_samples_partition) * (max_point_size - min_point_size) + min_point_size
                for size in class_counts
            ]
            plt.scatter([i] * self.num_classes, range(self.num_classes), s=sizes, alpha=0.5)

        plt.xlabel("Participant")
        plt.ylabel("Class")
        plt.xticks(range(self.partitions_number))
        plt.yticks(range(self.num_classes))
        if self.iid:
            plt.title(f"Participant {i + 1} class distribution (IID)")
        else:
            plt.title(
                f"Participant {i + 1} class distribution (Non-IID - {self.partition}) - {self.partition_parameter}"
            )
        plt.tight_layout()

        # Saves the distribution display with circles of different size
        path_to_save = f"{self.config.participant['tracking_args']['log_dir']}/{self.config.participant['scenario_args']['name']}/class_distribution_{'iid' if self.iid else 'non_iid'}{'_' + self.partition if not self.iid else ''}.png"
        plt.savefig(path_to_save, dpi=300, bbox_inches="tight")
        plt.close()

        if hasattr(self, "tsne") and self.tsne:
            self.visualize_tsne(dataset)

    def visualize_tsne(self, dataset):
        X = []  # List for storing the characteristics of the samples
        y = []  # Ready to store the labels of the samples
        for idx in range(len(dataset)):  # Assuming that 'dataset' is a list or array of your samples
            sample, label = dataset[idx]
            X.append(sample.flatten())
            y.append(label)

        X = np.array(X)
        y = np.array(y)

        tsne = TSNE(n_components=2, verbose=1, perplexity=40, n_iter=300)
        tsne_results = tsne.fit_transform(X)

        plt.figure(figsize=(16, 10))
        sns.scatterplot(
            x=tsne_results[:, 0],
            y=tsne_results[:, 1],
            hue=y,
            palette=sns.color_palette("hsv", self.num_classes),
            legend="full",
            alpha=0.7,
        )

        plt.title("t-SNE visualization of the dataset")
        plt.xlabel("t-SNE axis 1")
        plt.ylabel("t-SNE axis 2")
        plt.legend(title="Class")
        plt.tight_layout()

        path_to_save_tsne = f"{self.config.participant['tracking_args']['log_dir']}/{self.config.participant['scenario_args']['name']}/tsne_visualization.png"
        plt.savefig(path_to_save_tsne, dpi=300, bbox_inches="tight")
        plt.close()

    def dirichlet_partition(self, dataset, alpha=0.5, min_samples_per_class=10):
        y_data = self._get_targets(dataset)
        unique_labels = np.unique(y_data)
        logging_training.info(f"Labels unique: {unique_labels}")
        num_samples = len(y_data)

        indices_per_partition = [[] for _ in range(self.partitions_number)]
        label_distribution = self.class_distribution if self.class_distribution is not None else None

        for label in unique_labels:
            label_indices = np.where(y_data == label)[0]
            np.random.shuffle(label_indices)

            if label_distribution is None:
                proportions = np.random.dirichlet([alpha] * self.partitions_number)
            else:
                proportions = label_distribution[label]

            proportions = self._adjust_proportions(proportions, indices_per_partition, num_samples)
            split_points = (np.cumsum(proportions) * len(label_indices)).astype(int)[:-1]

            for partition_idx, indices in enumerate(np.split(label_indices, split_points)):
                if len(indices) < min_samples_per_class:
                    indices_per_partition[partition_idx].extend([])
                else:
                    indices_per_partition[partition_idx].extend(indices)

        if label_distribution is None:
            self.class_distribution = self._calculate_class_distribution(indices_per_partition, y_data)

        return {i: indices for i, indices in enumerate(indices_per_partition)}

    def _adjust_proportions(self, proportions, indices_per_partition, num_samples):
        adjusted = np.array([
            p * (len(indices) < num_samples / self.partitions_number)
            for p, indices in zip(proportions, indices_per_partition, strict=False)
        ])
        return adjusted / adjusted.sum()

    def _calculate_class_distribution(self, indices_per_partition, y_data):
        distribution = defaultdict(lambda: np.zeros(self.partitions_number))
        for partition_idx, indices in enumerate(indices_per_partition):
            labels, counts = np.unique(y_data[indices], return_counts=True)
            for label, count in zip(labels, counts, strict=False):
                distribution[label][partition_idx] = count
        return {k: v / v.sum() for k, v in distribution.items()}

    @staticmethod
    def _get_targets(dataset) -> np.ndarray:
        if isinstance(dataset.targets, np.ndarray):
            return dataset.targets
        elif hasattr(dataset.targets, "numpy"):
            return dataset.targets.numpy()
        else:
            return np.asarray(dataset.targets)

    def homo_partition(self, dataset):
        """
        Homogeneously partition the dataset into multiple subsets.

        This function divides a dataset into a specified number of subsets, where each subset
        is intended to have a roughly equal number of samples. This method aims to ensure a
        homogeneous distribution of data across all subsets. It's particularly useful in
        scenarios where a uniform distribution of data is desired among all federated learning
        clients.

        Args:
            dataset (torch.utils.data.Dataset): The dataset to partition. It should have
                                                'data' and 'targets' attributes.

        Returns:
            dict: A dictionary where keys are subset indices (ranging from 0 to partitions_number-1)
                and values are lists of indices corresponding to the samples in each subset.

        The function randomly shuffles the entire dataset and then splits it into the number
        of subsets specified by `partitions_number`. It ensures that each subset has a similar number
        of samples. The function also prints the class distribution in each subset for reference.

        Example usage:
            federated_data = homo_partition(my_dataset)
            # This creates federated data subsets with homogeneous distribution.
        """
        n_nets = self.partitions_number

        n_train = len(dataset.targets)
        np.random.seed(self.seed)
        idxs = np.random.permutation(n_train)
        batch_idxs = np.array_split(idxs, n_nets)
        net_dataidx_map = {i: batch_idxs[i] for i in range(n_nets)}

        # partitioned_datasets = []
        for i in range(self.partitions_number):
            # subset = torch.utils.data.Subset(dataset, net_dataidx_map[i])
            # partitioned_datasets.append(subset)

            # Print class distribution in the current partition
            class_counts = [0] * self.num_classes
            for idx in net_dataidx_map[i]:
                label = dataset.targets[idx]
                class_counts[label] += 1
            logging_training.info(f"Partition {i + 1} class distribution: {class_counts}")

        return net_dataidx_map

    def balanced_iid_partition(self, dataset):
        """
        Partition the dataset into balanced and IID (Independent and Identically Distributed)
        subsets for each client.

        This function divides a dataset into a specified number of subsets (federated clients),
        where each subset has an equal class distribution. This makes the partition suitable for
        simulating IID data scenarios in federated learning.

        Args:
            dataset (list): The dataset to partition. It should be a list of tuples where each
                            tuple represents a data sample and its corresponding label.

        Returns:
            dict: A dictionary where keys are client IDs (ranging from 0 to partitions_number-1) and
                    values are lists of indices corresponding to the samples assigned to each client.

        The function ensures that each class is represented equally in each subset. The
        partitioning process involves iterating over each class, shuffling the indices of that class,
        and then splitting them equally among the clients. The function does not print the class
        distribution in each subset.

        Example usage:
            federated_data = balanced_iid_partition(my_dataset)
            # This creates federated data subsets with equal class distributions.
        """
        num_clients = self.partitions_number
        clients_data = {i: [] for i in range(num_clients)}

        # Get the labels from the dataset
        if isinstance(dataset.targets, np.ndarray):
            labels = dataset.targets
        elif hasattr(dataset.targets, "numpy"):  # Check if it's a tensor with .numpy() method
            labels = dataset.targets.numpy()
        else:  # If it's a list
            labels = np.asarray(dataset.targets)

        label_counts = np.bincount(labels)
        min_label = label_counts.argmin()
        min_count = label_counts[min_label]

        for label in range(self.num_classes):
            # Get the indices of the same label samples
            label_indices = np.where(labels == label)[0]
            np.random.seed(self.seed)
            np.random.shuffle(label_indices)

            # Split the data based on their labels
            samples_per_client = min_count // num_clients

            for i in range(num_clients):
                start_idx = i * samples_per_client
                end_idx = (i + 1) * samples_per_client
                clients_data[i].extend(label_indices[start_idx:end_idx])

        return clients_data

    def unbalanced_iid_partition(self, dataset, imbalance_factor=2):
        """
        Partition the dataset into multiple IID (Independent and Identically Distributed)
        subsets with different size.

        This function divides a dataset into a specified number of IID subsets (federated
        clients), where each subset has a different number of samples. The number of samples
        in each subset is determined by an imbalance factor, making the partition suitable
        for simulating imbalanced data scenarios in federated learning.

        Args:
            dataset (list): The dataset to partition. It should be a list of tuples where
                            each tuple represents a data sample and its corresponding label.
            imbalance_factor (float): The factor to determine the degree of imbalance
                                    among the subsets. A lower imbalance factor leads to more
                                    imbalanced partitions.

        Returns:
            dict: A dictionary where keys are client IDs (ranging from 0 to partitions_number-1) and
                    values are lists of indices corresponding to the samples assigned to each client.

        The function ensures that each class is represented in each subset but with varying
        proportions. The partitioning process involves iterating over each class, shuffling
        the indices of that class, and then splitting them according to the calculated subset
        sizes. The function does not print the class distribution in each subset.

        Example usage:
            federated_data = unbalanced_iid_partition(my_dataset, imbalance_factor=2)
            # This creates federated data subsets with varying number of samples based on
            # an imbalance factor of 2.
        """
        num_clients = self.partitions_number
        clients_data = {i: [] for i in range(num_clients)}

        # Get the labels from the dataset
        labels = np.array([dataset.targets[idx] for idx in range(len(dataset))])
        label_counts = np.bincount(labels)

        min_label = label_counts.argmin()
        min_count = label_counts[min_label]

        # Set the initial_subset_size
        initial_subset_size = min_count // num_clients

        # Calculate the number of samples for each subset based on the imbalance factor
        subset_sizes = [initial_subset_size]
        for i in range(1, num_clients):
            subset_sizes.append(int(subset_sizes[i - 1] * ((imbalance_factor - 1) / imbalance_factor)))

        for label in range(self.num_classes):
            # Get the indices of the same label samples
            label_indices = np.where(labels == label)[0]
            np.random.seed(self.seed)
            np.random.shuffle(label_indices)

            # Split the data based on their labels
            start = 0
            for i in range(num_clients):
                end = start + subset_sizes[i]
                clients_data[i].extend(label_indices[start:end])
                start = end

        return clients_data

    def percentage_partition(self, dataset, percentage=20):
        """
        Partition a dataset into multiple subsets with a specified level of non-IID-ness.

        This function divides a dataset into a specified number of subsets (federated
        clients), where each subset has a different class distribution. The class
        distribution in each subset is determined by a specified percentage, making the
        partition suitable for simulating non-IID (non-Independently and Identically
        Distributed) data scenarios in federated learning.

        Args:
            dataset (torch.utils.data.Dataset): The dataset to partition. It should have
                                                'data' and 'targets' attributes.
            percentage (int): A value between 0 and 100 that specifies the desired
                                level of non-IID-ness for the labels of the federated data.
                                This percentage controls the imbalance in the class distribution
                                across different subsets.

        Returns:
            dict: A dictionary where keys are subset indices (ranging from 0 to partitions_number-1)
                and values are lists of indices corresponding to the samples in each subset.

        The function ensures that the number of classes in each subset varies based on the selected
        percentage. The partitioning process involves iterating over each class, shuffling the
        indices of that class, and then splitting them according to the calculated subset sizes.
        The function also prints the class distribution in each subset for reference.

        Example usage:
            federated_data = percentage_partition(my_dataset, percentage=20)
            # This creates federated data subsets with varying class distributions based on
            # a percentage of 20.
        """
        if isinstance(dataset.targets, np.ndarray):
            y_train = dataset.targets
        elif hasattr(dataset.targets, "numpy"):  # Check if it's a tensor with .numpy() method
            y_train = dataset.targets.numpy()
        else:  # If it's a list
            y_train = np.asarray(dataset.targets)

        num_classes = self.num_classes
        num_subsets = self.partitions_number
        class_indices = {i: np.where(y_train == i)[0] for i in range(num_classes)}

        # Get the labels from the dataset
        labels = np.array([dataset.targets[idx] for idx in range(len(dataset))])
        label_counts = np.bincount(labels)

        min_label = label_counts.argmin()
        min_count = label_counts[min_label]

        classes_per_subset = int(num_classes * percentage / 100)
        if classes_per_subset < 1:
            raise ValueError("The percentage is too low to assign at least one class to each subset.")

        subset_indices = [[] for _ in range(num_subsets)]
        class_list = list(range(num_classes))
        np.random.seed(self.seed)
        np.random.shuffle(class_list)

        for i in range(num_subsets):
            for j in range(classes_per_subset):
                # Use modulo operation to cycle through the class_list
                class_idx = class_list[(i * classes_per_subset + j) % num_classes]
                indices = class_indices[class_idx]
                np.random.seed(self.seed)
                np.random.shuffle(indices)
                # Select approximately 50% of the indices
                subset_indices[i].extend(indices[: min_count // 2])

            class_counts = np.bincount(np.array([dataset.targets[idx] for idx in subset_indices[i]]))
            logging_training.info(f"Partition {i + 1} class distribution: {class_counts.tolist()}")

        partitioned_datasets = {i: subset_indices[i] for i in range(num_subsets)}

        return partitioned_datasets

    def plot_all_data_distribution(self, dataset, partitions_map):
        """

        Plot all of the data distribution of the dataset according to the partitions map provided.

        Args:
            dataset: The dataset to plot (torch.utils.data.Dataset).
            partitions_map: The map of the dataset partitions.
        """
        sns.set()
        sns.set_style("whitegrid", {"axes.grid": False})
        sns.set_context("paper", font_scale=1.5)
        sns.set_palette("Set2")

        num_clients = len(partitions_map)
        num_classes = self.num_classes

        plt.figure(figsize=(12, 8))

        label_distribution = [[] for _ in range(num_classes)]
        for c_id, idc in partitions_map.items():
            for idx in idc:
                label_distribution[dataset.targets[idx]].append(c_id)

        plt.hist(
            label_distribution,
            stacked=True,
            bins=np.arange(-0.5, num_clients + 1.5, 1),
            label=dataset.classes,
            rwidth=0.5,
        )
        plt.xticks(
            np.arange(num_clients),
            ["Participant %d" % (c_id + 1) for c_id in range(num_clients)],
        )
        plt.title("Distribution of splited datasets")
        plt.xlabel("Participant")
        plt.ylabel("Number of samples")
        plt.xticks(range(num_clients), [f" {i}" for i in range(num_clients)])
        plt.legend(loc="upper right")
        plt.tight_layout()

        path_to_save = f"{self.config.participant['tracking_args']['log_dir']}/{self.config.participant['scenario_args']['name']}/all_data_distribution_{'iid' if self.iid else 'non_iid'}{'_' + self.partition if not self.iid else ''}.png"
        plt.savefig(path_to_save, dpi=300, bbox_inches="tight")
        plt.close()



import datetime
import ipaddress
import os

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization import Encoding, NoEncryption, PrivateFormat, load_pem_private_key
from cryptography.x509.oid import NameOID


def generate_ca_certificate(dir_path):
    keyfile_path = os.path.join(dir_path, "ca_key.pem")
    certfile_path = os.path.join(dir_path, "ca_cert.pem")

    if os.path.exists(keyfile_path) and os.path.exists(certfile_path):
        print("CA Certificate and key already exist")
        return keyfile_path, certfile_path

    # Generate certfile and keyfile for the CA (pem format)
    ca_private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048, backend=default_backend())

    ca_issuer = x509.Name([
        x509.NameAttribute(NameOID.COUNTRY_NAME, "ES"),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "Spain"),
        x509.NameAttribute(NameOID.LOCALITY_NAME, "Murcia"),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Nebula"),
        x509.NameAttribute(NameOID.COMMON_NAME, "ca.nebula"),
    ])

    valid_from = datetime.datetime.utcnow()
    valid_to = valid_from + datetime.timedelta(days=365)

    cert = (
        x509.CertificateBuilder()
        .subject_name(ca_issuer)
        .issuer_name(ca_issuer)
        .public_key(ca_private_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(valid_from)
        .not_valid_after(valid_to)
        .add_extension(x509.BasicConstraints(ca=True, path_length=None), critical=True)
        .sign(ca_private_key, hashes.SHA256(), default_backend())
    )

    with open(keyfile_path, "wb") as f:
        f.write(
            ca_private_key.private_bytes(
                encoding=Encoding.PEM,
                format=PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=NoEncryption(),
            )
        )

    with open(certfile_path, "wb") as f:
        f.write(cert.public_bytes(Encoding.PEM))

    return keyfile_path, certfile_path


def generate_certificate(dir_path, node_id, ip):
    keyfile_path = os.path.join(dir_path, f"{node_id}_key.pem")
    certfile_path = os.path.join(dir_path, f"{node_id}_cert.pem")
    ip_obj = ipaddress.ip_address(ip)

    if os.path.exists(keyfile_path) and os.path.exists(certfile_path):
        print("Certificate and key already exist")
        return keyfile_path, certfile_path

    with open(os.path.join(dir_path, "ca_key.pem"), "rb") as f:
        ca_private_key = load_pem_private_key(f.read(), password=None)

    with open(os.path.join(dir_path, "ca_cert.pem"), "rb") as f:
        ca_cert = x509.load_pem_x509_certificate(f.read())

    # Generate certfile and keyfile for the participant to use in the federation (pem format)
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048, backend=default_backend())

    subject = x509.Name([
        x509.NameAttribute(NameOID.COUNTRY_NAME, "US"),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "California"),
        x509.NameAttribute(NameOID.LOCALITY_NAME, "San Francisco"),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Nebula"),
        x509.NameAttribute(NameOID.COMMON_NAME, f"{node_id}.nebula"),
    ])

    valid_from = datetime.datetime.utcnow()
    valid_to = valid_from + datetime.timedelta(days=365)

    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(ca_cert.subject)
        .public_key(private_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(valid_from)
        .not_valid_after(valid_to)
        .add_extension(
            x509.SubjectAlternativeName([x509.DNSName("localhost"), x509.IPAddress(ip_obj)]),
            critical=False,
        )
        .sign(ca_private_key, hashes.SHA256(), default_backend())
    )

    with open(keyfile_path, "wb") as f:
        f.write(
            private_key.private_bytes(
                encoding=Encoding.PEM,
                format=PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=NoEncryption(),
            )
        )

    with open(certfile_path, "wb") as f:
        f.write(cert.public_bytes(Encoding.PEM))

    return keyfile_path, certfile_path


if __name__ == "__main__":
    current_dir = os.getcwd()
    generate_certificate(os.path.join(current_dir, "certs"))



import asyncio
import logging


async def debug_tasks():
    while True:
        tasks = asyncio.all_tasks()
        logging.info(f"Active tasks: {len(tasks)}")
        for task in tasks:
            logging.info(f"Task: {task}")
        await asyncio.sleep(5)



import asyncio
import logging
import os

import docker

from nebula.addons.attacks.attacks import create_attack
from nebula.addons.functions import print_msg_box
from nebula.addons.reporter import Reporter
from nebula.core.aggregation.aggregator import create_aggregator, create_malicious_aggregator, create_target_aggregator
from nebula.core.eventmanager import EventManager, event_handler
from nebula.core.network.communications import CommunicationsManager
from nebula.core.pb import nebula_pb2
from nebula.core.utils.locker import Locker

logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("fsspec").setLevel(logging.WARNING)
logging.getLogger("matplotlib").setLevel(logging.ERROR)
logging.getLogger("aim").setLevel(logging.ERROR)
logging.getLogger("plotly").setLevel(logging.ERROR)

import pdb
import sys

from nebula.config.config import Config
from nebula.core.training.lightning import Lightning
from nebula.core.utils.helper import cosine_metric


def handle_exception(exc_type, exc_value, exc_traceback):
    logging.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return
    pdb.set_trace()
    pdb.post_mortem(exc_traceback)


def signal_handler(sig, frame):
    print("Signal handler called with signal", sig)
    print("Exiting gracefully")
    sys.exit(0)


def print_banner():
    banner = """
                    ███╗   ██╗███████╗██████╗ ██╗   ██╗██╗      █████╗
                    ████╗  ██║██╔════╝██╔══██╗██║   ██║██║     ██╔══██╗
                    ██╔██╗ ██║█████╗  ██████╔╝██║   ██║██║     ███████║
                    ██║╚██╗██║██╔══╝  ██╔══██╗██║   ██║██║     ██╔══██║
                    ██║ ╚████║███████╗██████╔╝╚██████╔╝███████╗██║  ██║
                    ╚═╝  ╚═══╝╚══════╝╚═════╝  ╚═════╝ ╚══════╝╚═╝  ╚═╝
                      A Platform for Decentralized Federated Learning
                        Created by Enrique Tomás Martínez Beltrán
                          https://github.com/CyberDataLab/nebula
                """
    logging.info(f"\n{banner}\n")


class Engine:
    def __init__(
        self,
        model,
        dataset,
        config=Config,
        trainer=Lightning,
        security=False,
        model_poisoning=False,
        poisoned_ratio=0,
        noise_type="gaussian",
    ):
        self.config = config
        self.idx = config.participant["device_args"]["idx"]
        self.experiment_name = config.participant["scenario_args"]["name"]
        self.ip = config.participant["network_args"]["ip"]
        self.port = config.participant["network_args"]["port"]
        self.addr = config.participant["network_args"]["addr"]
        self.role = config.participant["device_args"]["role"]
        self.name = config.participant["device_args"]["name"]
        self.docker_id = config.participant["device_args"]["docker_id"]
        self.client = docker.from_env()

        print_banner()

        print_msg_box(
            msg=f"Name {self.name}\nRole: {self.role}",
            indent=2,
            title="Node information",
        )

        self._trainer = None
        self._aggregator = None
        self.round = None
        self.total_rounds = None
        self.federation_nodes = set()
        self.initialized = False
        self.log_dir = os.path.join(config.participant["tracking_args"]["log_dir"], self.experiment_name)

        self.security = security
        self.model_poisoning = model_poisoning
        self.poisoned_ratio = poisoned_ratio
        self.noise_type = noise_type

        self._trainer = trainer(model, dataset, config=self.config)
        self._aggregator = create_aggregator(config=self.config, engine=self)

        self._secure_neighbors = []
        self._is_malicious = True if self.config.participant["adversarial_args"]["attacks"] != "No Attack" else False

        msg = f"Trainer: {self._trainer.__class__.__name__}"
        msg += f"\nDataset: {self.config.participant['data_args']['dataset']}"
        msg += f"\nIID: {self.config.participant['data_args']['iid']}"
        msg += f"\nModel: {model.__class__.__name__}"
        msg += f"\nAggregation algorithm: {self._aggregator.__class__.__name__}"
        msg += f"\nNode behavior: {'malicious' if self._is_malicious else 'benign'}"
        print_msg_box(msg=msg, indent=2, title="Scenario information")
        print_msg_box(
            msg=f"Logging type: {self._trainer.logger.__class__.__name__}",
            indent=2,
            title="Logging information",
        )

        self.with_reputation = self.config.participant["defense_args"]["with_reputation"]
        self.is_dynamic_topology = self.config.participant["defense_args"]["is_dynamic_topology"]
        self.is_dynamic_aggregation = self.config.participant["defense_args"]["is_dynamic_aggregation"]
        self.target_aggregation = (
            create_target_aggregator(config=self.config, engine=self) if self.is_dynamic_aggregation else None
        )
        msg = f"Reputation system: {self.with_reputation}\nDynamic topology: {self.is_dynamic_topology}\nDynamic aggregation: {self.is_dynamic_aggregation}"
        msg += (
            f"\nTarget aggregation: {self.target_aggregation.__class__.__name__}" if self.is_dynamic_aggregation else ""
        )
        print_msg_box(msg=msg, indent=2, title="Defense information")

        self.learning_cycle_lock = Locker(name="learning_cycle_lock", async_lock=True)
        self.federation_setup_lock = Locker(name="federation_setup_lock", async_lock=True)
        self.federation_ready_lock = Locker(name="federation_ready_lock", async_lock=True)
        self.round_lock = Locker(name="round_lock", async_lock=True)

        self.config.reload_config_file()

        self._cm = CommunicationsManager(engine=self)
        # Set the communication manager in the model (send messages from there)
        self.trainer.model.set_communication_manager(self._cm)

        self._reporter = Reporter(config=self.config, trainer=self.trainer, cm=self.cm)

        self._event_manager = EventManager(
            default_callbacks=[
                self._discovery_discover_callback,
                self._control_alive_callback,
                self._connection_connect_callback,
                self._connection_disconnect_callback,
                self._federation_ready_callback,
                self._start_federation_callback,
                self._federation_models_included_callback,
            ]
        )

        # Register additional callbacks
        self._event_manager.register_event(
            (
                nebula_pb2.FederationMessage,
                nebula_pb2.FederationMessage.Action.REPUTATION,
            ),
            self._reputation_callback,
        )
        # ... add more callbacks here

    @property
    def cm(self):
        return self._cm

    @property
    def reporter(self):
        return self._reporter

    @property
    def event_manager(self):
        return self._event_manager

    @property
    def aggregator(self):
        return self._aggregator

    def get_aggregator_type(self):
        return type(self.aggregator)

    @property
    def trainer(self):
        return self._trainer

    def get_addr(self):
        return self.addr

    def get_config(self):
        return self.config

    def get_federation_nodes(self):
        return self.federation_nodes

    def get_initialization_status(self):
        return self.initialized

    def set_initialization_status(self, status):
        self.initialized = status

    def get_round(self):
        return self.round

    def get_federation_ready_lock(self):
        return self.federation_ready_lock

    def get_federation_setup_lock(self):
        return self.federation_setup_lock

    def get_round_lock(self):
        return self.round_lock

    @event_handler(nebula_pb2.DiscoveryMessage, nebula_pb2.DiscoveryMessage.Action.DISCOVER)
    async def _discovery_discover_callback(self, source, message):
        logging.info(
            f"🔍  handle_discovery_message | Trigger | Received discovery message from {source} (network propagation)"
        )
        current_connections = await self.cm.get_addrs_current_connections(myself=True)
        if source not in current_connections:
            logging.info(f"🔍  handle_discovery_message | Trigger | Connecting to {source} indirectly")
            await self.cm.connect(source, direct=False)
        async with self.cm.get_connections_lock():
            if source in self.cm.connections:
                # Update the latitude and longitude of the node (if already connected)
                if (
                    message.latitude is not None
                    and -90 <= message.latitude <= 90
                    and message.longitude is not None
                    and -180 <= message.longitude <= 180
                ):
                    self.cm.connections[source].update_geolocation(message.latitude, message.longitude)
                else:
                    logging.warning(
                        f"🔍  Invalid geolocation received from {source}: latitude={message.latitude}, longitude={message.longitude}"
                    )

    @event_handler(nebula_pb2.ControlMessage, nebula_pb2.ControlMessage.Action.ALIVE)
    async def _control_alive_callback(self, source, message):
        logging.info(f"🔧  handle_control_message | Trigger | Received alive message from {source}")
        current_connections = await self.cm.get_addrs_current_connections(myself=True)
        if source in current_connections:
            try:
                await self.cm.health.alive(source)
            except Exception as e:
                logging.exception(f"Error updating alive status in connection: {e}")
        else:
            logging.error(f"❗️  Connection {source} not found in connections...")

    @event_handler(nebula_pb2.ConnectionMessage, nebula_pb2.ConnectionMessage.Action.CONNECT)
    async def _connection_connect_callback(self, source, message):
        logging.info(f"🔗  handle_connection_message | Trigger | Received connection message from {source}")
        current_connections = await self.cm.get_addrs_current_connections(myself=True)
        if source not in current_connections:
            logging.info(f"🔗  handle_connection_message | Trigger | Connecting to {source}")
            await self.cm.connect(source, direct=True)

    @event_handler(nebula_pb2.ConnectionMessage, nebula_pb2.ConnectionMessage.Action.DISCONNECT)
    async def _connection_disconnect_callback(self, source, message):
        logging.info(f"🔗  handle_connection_message | Trigger | Received disconnection message from {source}")
        await self.cm.disconnect(source, mutual_disconnection=False)

    @event_handler(
        nebula_pb2.FederationMessage,
        nebula_pb2.FederationMessage.Action.FEDERATION_READY,
    )
    async def _federation_ready_callback(self, source, message):
        logging.info(f"📝  handle_federation_message | Trigger | Received ready federation message from {source}")
        if self.config.participant["device_args"]["start"]:
            logging.info(f"📝  handle_federation_message | Trigger | Adding ready connection {source}")
            await self.cm.add_ready_connection(source)

    @event_handler(
        nebula_pb2.FederationMessage,
        nebula_pb2.FederationMessage.Action.FEDERATION_START,
    )
    async def _start_federation_callback(self, source, message):
        logging.info(f"📝  handle_federation_message | Trigger | Received start federation message from {source}")
        await self.create_trainer_module()

    @event_handler(nebula_pb2.FederationMessage, nebula_pb2.FederationMessage.Action.REPUTATION)
    async def _reputation_callback(self, source, message):
        malicious_nodes = message.arguments  # List of malicious nodes
        if self.with_reputation:
            if len(malicious_nodes) > 0 and not self._is_malicious:
                if self.is_dynamic_topology:
                    await self._disrupt_connection_using_reputation(malicious_nodes)
                if self.is_dynamic_aggregation and self.aggregator != self.target_aggregation:
                    await self._dynamic_aggregator(
                        self.aggregator.get_nodes_pending_models_to_aggregate(),
                        malicious_nodes,
                    )

    @event_handler(
        nebula_pb2.FederationMessage,
        nebula_pb2.FederationMessage.Action.FEDERATION_MODELS_INCLUDED,
    )
    async def _federation_models_included_callback(self, source, message):
        logging.info(f"📝  handle_federation_message | Trigger | Received aggregation finished message from {source}")
        try:
            await self.cm.get_connections_lock().acquire_async()
            if self.round is not None and source in self.cm.connections:
                try:
                    if message is not None and len(message.arguments) > 0:
                        self.cm.connections[source].update_round(int(message.arguments[0])) if message.round in [
                            self.round - 1,
                            self.round,
                        ] else None
                except Exception as e:
                    logging.exception(f"Error updating round in connection: {e}")
            else:
                logging.error(f"Connection not found for {source}")
        except Exception as e:
            logging.exception(f"Error updating round in connection: {e}")
        finally:
            await self.cm.get_connections_lock().release_async()

    async def create_trainer_module(self):
        asyncio.create_task(self._start_learning())
        logging.info("Started trainer module...")

    async def start_communications(self):
        logging.info(f"Neighbors: {self.config.participant['network_args']['neighbors']}")
        logging.info(
            f"💤  Cold start time: {self.config.participant['misc_args']['grace_time_connection']} seconds before connecting to the network"
        )
        await asyncio.sleep(self.config.participant["misc_args"]["grace_time_connection"])
        await self.cm.start()
        initial_neighbors = self.config.participant["network_args"]["neighbors"].split()
        for i in initial_neighbors:
            addr = f"{i.split(':')[0]}:{i.split(':')[1]}"
            await self.cm.connect(addr, direct=True)
            await asyncio.sleep(1)
        while not self.cm.verify_connections(initial_neighbors):
            await asyncio.sleep(1)
        current_connections = await self.cm.get_addrs_current_connections()
        logging.info(f"Connections verified: {current_connections}")
        await self._reporter.start()
        await self.cm.deploy_additional_services()
        await asyncio.sleep(self.config.participant["misc_args"]["grace_time_connection"] // 2)

    async def deploy_federation(self):
        await self.federation_ready_lock.acquire_async()
        if self.config.participant["device_args"]["start"]:
            logging.info(
                f"💤  Waiting for {self.config.participant['misc_args']['grace_time_start_federation']} seconds to start the federation"
            )
            await asyncio.sleep(self.config.participant["misc_args"]["grace_time_start_federation"])
            if self.round is None:
                while not await self.cm.check_federation_ready():
                    await asyncio.sleep(1)
                logging.info("Sending FEDERATION_START to neighbors...")
                message = self.cm.mm.generate_federation_message(nebula_pb2.FederationMessage.Action.FEDERATION_START)
                await self.cm.send_message_to_neighbors(message)
                await self.get_federation_ready_lock().release_async()
                await self.create_trainer_module()
            else:
                logging.info("Federation already started")

        else:
            logging.info("Sending FEDERATION_READY to neighbors...")
            message = self.cm.mm.generate_federation_message(nebula_pb2.FederationMessage.Action.FEDERATION_READY)
            await self.cm.send_message_to_neighbors(message)
            logging.info("💤  Waiting until receiving the start signal from the start node")

    async def _start_learning(self):
        await self.learning_cycle_lock.acquire_async()
        try:
            if self.round is None:
                self.total_rounds = self.config.participant["scenario_args"]["rounds"]
                epochs = self.config.participant["training_args"]["epochs"]
                await self.get_round_lock().acquire_async()
                self.round = 0
                await self.get_round_lock().release_async()
                await self.learning_cycle_lock.release_async()
                print_msg_box(
                    msg="Starting Federated Learning process...",
                    indent=2,
                    title="Start of the experiment",
                )
                direct_connections = await self.cm.get_addrs_current_connections(only_direct=True)
                undirected_connections = await self.cm.get_addrs_current_connections(only_undirected=True)
                logging.info(
                    f"Initial DIRECT connections: {direct_connections} | Initial UNDIRECT participants: {undirected_connections}"
                )
                logging.info("💤  Waiting initialization of the federation...")
                # Lock to wait for the federation to be ready (only affects the first round, when the learning starts)
                # Only applies to non-start nodes --> start node does not wait for the federation to be ready
                await self.get_federation_ready_lock().acquire_async()
                if self.config.participant["device_args"]["start"]:
                    logging.info("Propagate initial model updates.")
                    await self.cm.propagator.propagate("initialization")
                    await self.get_federation_ready_lock().release_async()

                self.trainer.set_epochs(epochs)
                self.trainer.create_trainer()

                await self._learning_cycle()
            else:
                if await self.learning_cycle_lock.locked_async():
                    await self.learning_cycle_lock.release_async()
        finally:
            if await self.learning_cycle_lock.locked_async():
                await self.learning_cycle_lock.release_async()

    async def _disrupt_connection_using_reputation(self, malicious_nodes):
        malicious_nodes = list(set(malicious_nodes) & set(self.get_current_connections()))
        logging.info(f"Disrupting connection with malicious nodes at round {self.round}")
        logging.info(f"Removing {malicious_nodes} from {self.get_current_connections()}")
        logging.info(f"Current connections before aggregation at round {self.round}: {self.get_current_connections()}")
        for malicious_node in malicious_nodes:
            if (self.get_name() != malicious_node) and (malicious_node not in self._secure_neighbors):
                await self.cm.disconnect(malicious_node)
        logging.info(f"Current connections after aggregation at round {self.round}: {self.get_current_connections()}")

        await self._connect_with_benign(malicious_nodes)

    async def _connect_with_benign(self, malicious_nodes):
        lower_threshold = 1
        higher_threshold = len(self.federation_nodes) - 1
        if higher_threshold < lower_threshold:
            higher_threshold = lower_threshold

        benign_nodes = [i for i in self.federation_nodes if i not in malicious_nodes]
        logging.info(f"_reputation_callback benign_nodes at round {self.round}: {benign_nodes}")
        if len(self.get_current_connections()) <= lower_threshold:
            for node in benign_nodes:
                if len(self.get_current_connections()) <= higher_threshold and self.get_name() != node:
                    connected = await self.cm.connect(node)
                    if connected:
                        logging.info(f"Connect new connection with at round {self.round}: {connected}")

    async def _dynamic_aggregator(self, aggregated_models_weights, malicious_nodes):
        logging.info(f"malicious detected at round {self.round}, change aggergation protocol!")
        if self.aggregator != self.target_aggregation:
            logging.info(f"Current aggregator is: {self.aggregator}")
            self.aggregator = self.target_aggregation
            await self.aggregator.update_federation_nodes(self.federation_nodes)

            for subnodes in aggregated_models_weights.keys():
                sublist = subnodes.split()
                (submodel, weights) = aggregated_models_weights[subnodes]
                for node in sublist:
                    if node not in malicious_nodes:
                        await self.aggregator.include_model_in_buffer(
                            submodel, weights, source=self.get_name(), round=self.round
                        )
            logging.info(f"Current aggregator is: {self.aggregator}")

    async def _waiting_model_updates(self):
        logging.info(f"💤  Waiting convergence in round {self.round}.")
        params = await self.aggregator.get_aggregation()
        if params is not None:
            logging.info(
                f"_waiting_model_updates | Aggregation done for round {self.round}, including parameters in local model."
            )
            self.trainer.set_model_parameters(params)
        else:
            logging.error("Aggregation finished with no parameters")

    async def _learning_cycle(self):
        while self.round is not None and self.round < self.total_rounds:
            print_msg_box(
                msg=f"Round {self.round} of {self.total_rounds} started.",
                indent=2,
                title="Round information",
            )
            self.trainer.on_round_start()
            self.federation_nodes = await self.cm.get_addrs_current_connections(only_direct=True, myself=True)
            logging.info(f"Federation nodes: {self.federation_nodes}")
            direct_connections = await self.cm.get_addrs_current_connections(only_direct=True)
            undirected_connections = await self.cm.get_addrs_current_connections(only_undirected=True)
            logging.info(f"Direct connections: {direct_connections} | Undirected connections: {undirected_connections}")
            logging.info(f"[Role {self.role}] Starting learning cycle...")
            await self.aggregator.update_federation_nodes(self.federation_nodes)
            await self._extended_learning_cycle()

            await self.get_round_lock().acquire_async()
            print_msg_box(
                msg=f"Round {self.round} of {self.total_rounds} finished.",
                indent=2,
                title="Round information",
            )
            await self.aggregator.reset()
            self.trainer.on_round_end()
            self.round = self.round + 1
            self.config.participant["federation_args"]["round"] = (
                self.round
            )  # Set current round in config (send to the controller)
            await self.get_round_lock().release_async()

        # End of the learning cycle
        self.trainer.on_learning_cycle_end()
        await self.trainer.test()
        self.round = None
        self.total_rounds = None
        print_msg_box(
            msg="Federated Learning process has been completed.",
            indent=2,
            title="End of the experiment",
        )
        # Report
        if self.config.participant["scenario_args"]["controller"] != "nebula-test":
            result = await self.reporter.report_scenario_finished()
            if result:
                pass
            else:
                logging.error("Error reporting scenario finished")

        logging.info("Checking if all my connections reached the total rounds...")
        while not self.cm.check_finished_experiment():
            await asyncio.sleep(1)

        # Enable loggin info
        logging.getLogger().disabled = True

        # Kill itself
        if self.config.participant["scenario_args"]["deployment"] == "docker":
            try:
                self.client.containers.get(self.docker_id).stop()
            except Exception as e:
                print(f"Error stopping Docker container with ID {self.docker_id}: {e}")

    async def _extended_learning_cycle(self):
        """
        This method is called in each round of the learning cycle. It is used to extend the learning cycle with additional
        functionalities. The method is called in the _learning_cycle method.
        """
        pass

    def reputation_calculation(self, aggregated_models_weights):
        cossim_threshold = 0.5
        loss_threshold = 0.5

        current_models = {}
        for subnodes in aggregated_models_weights.keys():
            sublist = subnodes.split()
            submodel = aggregated_models_weights[subnodes][0]
            for node in sublist:
                current_models[node] = submodel

        malicious_nodes = []
        reputation_score = {}
        local_model = self.trainer.get_model_parameters()
        untrusted_nodes = list(current_models.keys())
        logging.info(f"reputation_calculation untrusted_nodes at round {self.round}: {untrusted_nodes}")

        for untrusted_node in untrusted_nodes:
            logging.info(f"reputation_calculation untrusted_node at round {self.round}: {untrusted_node}")
            logging.info(f"reputation_calculation self.get_name() at round {self.round}: {self.get_name()}")
            if untrusted_node != self.get_name():
                untrusted_model = current_models[untrusted_node]
                cossim = cosine_metric(local_model, untrusted_model, similarity=True)
                logging.info(f"reputation_calculation cossim at round {self.round}: {untrusted_node}: {cossim}")
                self.trainer._logger.log_data({f"Reputation/cossim_{untrusted_node}": cossim}, step=self.round)

                avg_loss = self.trainer.validate_neighbour_model(untrusted_model)
                logging.info(f"reputation_calculation avg_loss at round {self.round} {untrusted_node}: {avg_loss}")
                self.trainer._logger.log_data({f"Reputation/avg_loss_{untrusted_node}": avg_loss}, step=self.round)
                reputation_score[untrusted_node] = (cossim, avg_loss)

                if cossim < cossim_threshold or avg_loss > loss_threshold:
                    malicious_nodes.append(untrusted_node)
                else:
                    self._secure_neighbors.append(untrusted_node)

        return malicious_nodes, reputation_score

    async def send_reputation(self, malicious_nodes):
        logging.info(f"Sending REPUTATION to the rest of the topology: {malicious_nodes}")
        message = self.cm.mm.generate_federation_message(
            nebula_pb2.FederationMessage.Action.REPUTATION, malicious_nodes
        )
        await self.cm.send_message_to_neighbors(message)


class MaliciousNode(Engine):
    def __init__(
        self,
        model,
        dataset,
        config=Config,
        trainer=Lightning,
        security=False,
        model_poisoning=False,
        poisoned_ratio=0,
        noise_type="gaussian",
    ):
        super().__init__(
            model,
            dataset,
            config,
            trainer,
            security,
            model_poisoning,
            poisoned_ratio,
            noise_type,
        )
        self.attack = create_attack(config.participant["adversarial_args"]["attacks"])
        self.fit_time = 0.0
        self.extra_time = 0.0

        self.round_start_attack = 3
        self.round_stop_attack = 6

        self.aggregator_bening = self._aggregator

    async def _extended_learning_cycle(self):
        if self.attack != None:
            if self.round in range(self.round_start_attack, self.round_stop_attack):
                logging.info("Changing aggregation function maliciously...")
                self._aggregator = create_malicious_aggregator(self._aggregator, self.attack)
            elif self.round == self.round_stop_attack:
                logging.info("Changing aggregation function benignly...")
                self._aggregator = self.aggregator_bening

        if self.role == "aggregator":
            await AggregatorNode._extended_learning_cycle(self)
        if self.role == "trainer":
            await TrainerNode._extended_learning_cycle(self)
        if self.role == "server":
            await ServerNode._extended_learning_cycle(self)


class AggregatorNode(Engine):
    def __init__(
        self,
        model,
        dataset,
        config=Config,
        trainer=Lightning,
        security=False,
        model_poisoning=False,
        poisoned_ratio=0,
        noise_type="gaussian",
    ):
        super().__init__(
            model,
            dataset,
            config,
            trainer,
            security,
            model_poisoning,
            poisoned_ratio,
            noise_type,
        )

    async def _extended_learning_cycle(self):
        # Define the functionality of the aggregator node
        await self.trainer.test()
        await self.trainer.train()

        await self.aggregator.include_model_in_buffer(
            self.trainer.get_model_parameters(),
            self.trainer.get_model_weight(),
            source=self.addr,
            round=self.round,
        )

        await self.cm.propagator.propagate("stable")
        await self._waiting_model_updates()


class ServerNode(Engine):
    def __init__(
        self,
        model,
        dataset,
        config=Config,
        trainer=Lightning,
        security=False,
        model_poisoning=False,
        poisoned_ratio=0,
        noise_type="gaussian",
    ):
        super().__init__(
            model,
            dataset,
            config,
            trainer,
            security,
            model_poisoning,
            poisoned_ratio,
            noise_type,
        )

    async def _extended_learning_cycle(self):
        # Define the functionality of the server node
        await self.trainer.test()

        # In the first round, the server node doest take into account the initial model parameters for the aggregation
        await self.aggregator.include_model_in_buffer(
            self.trainer.get_model_parameters(),
            self.trainer.BYPASS_MODEL_WEIGHT,
            source=self.addr,
            round=self.round,
        )
        await self._waiting_model_updates()
        await self.cm.propagator.propagate("stable")


class TrainerNode(Engine):
    def __init__(
        self,
        model,
        dataset,
        config=Config,
        trainer=Lightning,
        security=False,
        model_poisoning=False,
        poisoned_ratio=0,
        noise_type="gaussian",
    ):
        super().__init__(
            model,
            dataset,
            config,
            trainer,
            security,
            model_poisoning,
            poisoned_ratio,
            noise_type,
        )

    async def _extended_learning_cycle(self):
        # Define the functionality of the trainer node
        logging.info("Waiting global update | Assign _waiting_global_update = True")
        self.aggregator.set_waiting_global_update()

        await self.trainer.test()
        await self.trainer.train()

        await self.aggregator.include_model_in_buffer(
            self.trainer.get_model_parameters(),
            self.trainer.get_model_weight(),
            source=self.addr,
            round=self.round,
            local=True,
        )

        await self.cm.propagator.propagate("stable")
        await self._waiting_model_updates()


class IdleNode(Engine):
    def __init__(
        self,
        model,
        dataset,
        config=Config,
        trainer=Lightning,
        security=False,
        model_poisoning=False,
        poisoned_ratio=0,
        noise_type="gaussian",
    ):
        super().__init__(
            model,
            dataset,
            config,
            trainer,
            security,
            model_poisoning,
            poisoned_ratio,
            noise_type,
        )

    async def _extended_learning_cycle(self):
        # Define the functionality of the idle node
        logging.info("Waiting global update | Assign _waiting_global_update = True")
        self.aggregator.set_waiting_global_update()
        await self._waiting_model_updates()



import logging

import torch
from lightning import LightningDataModule
from torch.utils.data import DataLoader, RandomSampler, random_split

from nebula.config.config import TRAINING_LOGGER
from nebula.core.datasets.changeablesubset import ChangeableSubset

logging_training = logging.getLogger(TRAINING_LOGGER)


class DataModule(LightningDataModule):
    def __init__(
        self,
        train_set,
        train_set_indices,
        test_set,
        test_set_indices,
        local_test_set_indices,
        partition_id=0,
        partitions_number=1,
        batch_size=32,
        num_workers=0,
        val_percent=0.1,
        label_flipping=False,
        data_poisoning=False,
        poisoned_persent=0,
        poisoned_ratio=0,
        targeted=False,
        target_label=0,
        target_changed_label=0,
        noise_type="salt",
        seed=42,
    ):
        super().__init__()
        self.train_set = train_set
        self.train_set_indices = train_set_indices
        self.test_set = test_set
        self.test_set_indices = test_set_indices
        self.local_test_set_indices = local_test_set_indices
        self.partition_id = partition_id
        self.partitions_number = partitions_number
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.val_percent = val_percent
        self.label_flipping = label_flipping
        self.data_poisoning = data_poisoning
        self.poisoned_persent = poisoned_persent
        self.poisoned_ratio = poisoned_ratio
        self.targeted = targeted
        self.target_label = target_label
        self.target_changed_label = target_changed_label
        self.noise_type = noise_type
        self.seed = seed

        self.model_weight = None

        self.val_indices = None
        # Split train and validation datasets
        self.data_train = None
        self.data_val = None
        self.global_te_subset = None
        self.local_te_subset = None

    def setup(self, stage=None):
        if stage in (None, "fit"):
            tr_subset = ChangeableSubset(
                self.train_set,
                self.train_set_indices,
                label_flipping=self.label_flipping,
                data_poisoning=self.data_poisoning,
                poisoned_persent=self.poisoned_persent,
                poisoned_ratio=self.poisoned_ratio,
                targeted=self.targeted,
                target_label=self.target_label,
                target_changed_label=self.target_changed_label,
                noise_type=self.noise_type,
            )

            if self.val_indices is None:
                generator = torch.Generator()
                generator.manual_seed(self.seed)

                train_size = round(len(tr_subset) * (1 - self.val_percent))
                val_size = len(tr_subset) - train_size

                self.data_train, self.data_val = random_split(tr_subset, [train_size, val_size], generator=generator)

                self.val_indices = self.data_val.indices

            else:
                train_indices = list(set(range(len(tr_subset))) - set(self.val_indices))
                val_indices = self.val_indices

                self.data_train = ChangeableSubset(tr_subset, train_indices)
                self.data_val = ChangeableSubset(tr_subset, val_indices)

            self.model_weight = len(self.data_train)

        if stage in (None, "test"):
            # Test sets
            self.global_te_subset = ChangeableSubset(self.test_set, self.test_set_indices)
            self.local_te_subset = ChangeableSubset(self.test_set, self.local_test_set_indices)

            if len(self.test_set) < self.partitions_number:
                raise ValueError("Too many partitions for the size of the test set.")

    def teardown(self, stage=None):
        # Teardown the datasets
        if stage in (None, "fit"):
            self.data_train = None
            self.data_val = None

        if stage in (None, "test"):
            self.global_te_subset = None
            self.local_te_subset = None

    def train_dataloader(self):
        if self.data_train is None:
            raise ValueError(
                "Train dataset not initialized. Please call setup('fit') before requesting train_dataloader."
            )
        logging_training.info(f"Train set size: {len(self.data_train)}")
        return DataLoader(
            self.data_train,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.num_workers,
            drop_last=True,
            pin_memory=False,
        )

    def val_dataloader(self):
        if self.data_val is None:
            raise ValueError(
                "Validation dataset not initialized. Please call setup('fit') before requesting val_dataloader."
            )
        logging_training.info(f"Validation set size: {len(self.data_val)}")
        return DataLoader(
            self.data_val,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            drop_last=True,
            pin_memory=False,
        )

    def test_dataloader(self):
        if self.local_te_subset is None or self.global_te_subset is None:
            raise ValueError(
                "Test datasets not initialized. Please call setup('test') before requesting test_dataloader."
            )
        logging_training.info(f"Local test set size: {len(self.local_te_subset)}")
        logging_training.info(f"Global test set size: {len(self.global_te_subset)}")
        return [
            DataLoader(
                self.local_te_subset,
                batch_size=self.batch_size,
                shuffle=False,
                num_workers=self.num_workers,
                drop_last=True,
                pin_memory=False,
            ),
            DataLoader(
                self.global_te_subset,
                batch_size=self.batch_size,
                shuffle=False,
                num_workers=self.num_workers,
                drop_last=True,
                pin_memory=False,
            ),
        ]

    def bootstrap_dataloader(self):
        if self.data_val is None:
            raise ValueError(
                "Validation dataset not initialized. Please call setup('fit') before requesting bootstrap_dataloader."
            )
        random_sampler = RandomSampler(
            data_source=self.data_val,
            replacement=False,
            num_samples=max(int(len(self.data_val) / 3), 300),
        )
        logging_training.info(f"Bootstrap samples: {len(random_sampler)}")
        return DataLoader(
            self.data_train,
            batch_size=self.batch_size,
            shuffle=False,
            sampler=random_sampler,
            num_workers=self.num_workers,
            drop_last=True,
            pin_memory=False,
        )



import logging
from datetime import datetime

from lightning.pytorch.loggers import TensorBoardLogger


class NebulaTensorBoardLogger(TensorBoardLogger):
    def __init__(self, scenario_start_time, *args, **kwargs):
        self.scenario_start_time = scenario_start_time
        self.local_step = 0
        self.global_step = 0
        super().__init__(*args, **kwargs)

    def get_step(self):
        return int((datetime.now() - datetime.strptime(self.scenario_start_time, "%d/%m/%Y %H:%M:%S")).total_seconds())

    def log_data(self, data, step=None):
        if step is None:
            step = self.get_step()
        # logging.debug(f"Logging data for global step {step} | local step {self.local_step} | global step {self.global_step}")
        try:
            super().log_metrics(data, step)
        except ValueError:
            pass
        except Exception as e:
            logging.exception(f"Error logging statistics data [{data}] for step [{step}]: {e}")

    def log_metrics(self, metrics, step=None):
        if step is None:
            self.local_step += 1
            step = self.global_step + self.local_step
        # logging.debug(f"Logging metrics for global step {step} | local step {self.local_step} | global step {self.global_step}")
        if "epoch" in metrics:
            metrics.pop("epoch")
        try:
            super().log_metrics(metrics, step)
        except Exception as e:
            logging.exception(f"Error logging metrics [{metrics}] for step [{step}]: {e}")

    def log_figure(self, figure, step=None, name=None):
        if step is None:
            step = self.get_step()
        try:
            self.experiment.add_figure(name, figure, step)
        except Exception as e:
            logging.exception(f"Error logging figure [{name}] for step [{step}]: {e}")

    def get_logger_config(self):
        return {
            "scenario_start_time": self.scenario_start_time,
            "local_step": self.local_step,
            "global_step": self.global_step,
        }

    def set_logger_config(self, logger_config):
        if logger_config is None:
            return
        try:
            self.scenario_start_time = logger_config["scenario_start_time"]
            self.local_step = logger_config["local_step"]
            self.global_step = logger_config["global_step"]
        except Exception as e:
            logging.exception(f"Error setting logger config: {e}")



import torch

from nebula.core.models.nebulamodel import NebulaModel


class MNISTModelCNN(NebulaModel):
    def __init__(
        self,
        input_channels=1,
        num_classes=10,
        learning_rate=1e-3,
        metrics=None,
        confusion_matrix=None,
        seed=None,
    ):
        super().__init__(input_channels, num_classes, learning_rate, metrics, confusion_matrix, seed)

        self.config = {"beta1": 0.851436, "beta2": 0.999689, "amsgrad": True}

        self.example_input_array = torch.zeros(1, 1, 28, 28)
        self.learning_rate = learning_rate
        self.criterion = torch.nn.CrossEntropyLoss()
        self.conv1 = torch.nn.Conv2d(
            in_channels=input_channels,
            out_channels=32,
            kernel_size=(5, 5),
            padding="same",
        )
        self.relu = torch.nn.ReLU()
        self.pool1 = torch.nn.MaxPool2d(kernel_size=(2, 2), stride=2)
        self.conv2 = torch.nn.Conv2d(in_channels=32, out_channels=64, kernel_size=(5, 5), padding="same")
        self.pool2 = torch.nn.MaxPool2d(kernel_size=(2, 2), stride=2)
        self.l1 = torch.nn.Linear(7 * 7 * 64, 2048)
        self.l2 = torch.nn.Linear(2048, num_classes)

    def forward(self, x):
        input_layer = x.view(-1, 1, 28, 28)
        conv1 = self.relu(self.conv1(input_layer))
        pool1 = self.pool1(conv1)
        conv2 = self.relu(self.conv2(pool1))
        pool2 = self.pool2(conv2)
        pool2_flat = pool2.reshape(-1, 7 * 7 * 64)

        dense = self.relu(self.l1(pool2_flat))
        logits = self.l2(dense)
        return logits

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(
            self.parameters(),
            lr=self.learning_rate,
            betas=(self.config["beta1"], self.config["beta2"]),
            amsgrad=self.config["amsgrad"],
        )
        return optimizer



import logging

from nebula.addons.trustworthiness import calculation
from nebula.addons.trustworthiness.utils import get_input_value

logger = logging.getLogger(__name__)


class TrustPillar:
    """
    Class to represent a trust pillar.

    Args:
        name (string): Name of the pillar.
        metrics (dict): Metric definitions for the pillar.
        input_docs (dict): Input documents.
        use_weights (bool): True to turn on the weights in the metric config file.

    """

    def __init__(self, name, metrics, input_docs, use_weights=False):
        self.name = name
        self.input_docs = input_docs
        self.metrics = metrics
        self.result = []
        self.use_weights = use_weights

    def evaluate(self):
        """
        Evaluate the trust score for the pillar.

        Returns:
            float: Score of [0, 1].
        """
        score = 0
        avg_weight = 1 / len(self.metrics)
        for key, value in self.metrics.items():
            weight = value.get("weight", avg_weight) if self.use_weights else avg_weight
            score += weight * self.get_notion_score(key, value.get("metrics"))
        score = round(score, 2)
        return score, {self.name: {"score": score, "notions": self.result}}

    def get_notion_score(self, name, metrics):
        """
        Evaluate the trust score for the notion.

        Args:
            name (string): Name of the notion.
            metrics (list): Metrics definitions of the notion.

        Returns:
            float: Score of [0, 1].
        """

        notion_score = 0
        avg_weight = 1 / len(metrics)
        metrics_result = []
        for key, value in metrics.items():
            metric_score = self.get_metric_score(metrics_result, key, value)
            weight = value.get("weight", avg_weight) if self.use_weights else avg_weight
            notion_score += weight * float(metric_score)
        self.result.append({name: {"score": notion_score, "metrics": metrics_result}})
        return notion_score

    def get_metric_score(self, result, name, metric):
        """
        Evaluate the trust score for the metric.

        Args:
            result (object): The result object
            name (string): Name of the metric.
            metrics (dict): The metric definition.

        Returns:
            float: Score of [0, 1].
        """

        score = 0
        try:
            input_value = get_input_value(self.input_docs, metric.get("inputs"), metric.get("operation"))

            score_type = metric.get("type")
            if input_value is None:
                logger.warning(f"{name} input value is null")
            else:
                if score_type == "true_score":
                    score = calculation.get_true_score(input_value, metric.get("direction"))
                elif score_type == "score_mapping":
                    score = calculation.get_mapped_score(input_value, metric.get("score_map"))
                elif score_type == "ranges":
                    score = calculation.get_range_score(input_value, metric.get("ranges"), metric.get("direction"))
                elif score_type == "score_map_value":
                    score = calculation.get_map_value_score(input_value, metric.get("score_map"))
                elif score_type == "scaled_score":
                    score = calculation.get_scaled_score(input_value, metric.get("scale"), metric.get("direction"))
                elif score_type == "property_check":
                    score = 0 if input_value is None else input_value

                else:
                    logger.warning(f"The score type {score_type} is not yet implemented.")

        except KeyError:
            logger.warning(f"Null input for {name} metric")
        score = round(score, 2)
        result.append({name: {"score": score}})
        return score



from torch.utils.data import Subset

from nebula.addons.attacks.poisoning.datapoison import datapoison
from nebula.addons.attacks.poisoning.labelflipping import labelFlipping


class ChangeableSubset(Subset):
    def __init__(
        self,
        dataset,
        indices,
        label_flipping=False,
        data_poisoning=False,
        poisoned_persent=0,
        poisoned_ratio=0,
        targeted=False,
        target_label=0,
        target_changed_label=0,
        noise_type="salt",
    ):
        super().__init__(dataset, indices)
        new_dataset = dataset
        self.dataset = new_dataset
        self.indices = indices
        self.label_flipping = label_flipping
        self.data_poisoning = data_poisoning
        self.poisoned_persent = poisoned_persent
        self.poisoned_ratio = poisoned_ratio
        self.targeted = targeted
        self.target_label = target_label
        self.target_changed_label = target_changed_label
        self.noise_type = noise_type

        if self.label_flipping:
            self.dataset = labelFlipping(
                self.dataset,
                self.indices,
                self.poisoned_persent,
                self.targeted,
                self.target_label,
                self.target_changed_label,
            )
        if self.data_poisoning:
            self.dataset = datapoison(
                self.dataset,
                self.indices,
                self.poisoned_persent,
                self.poisoned_ratio,
                self.targeted,
                self.target_label,
                self.noise_type,
            )

    def __getitem__(self, idx):
        if isinstance(idx, list):
            return self.dataset[[self.indices[i] for i in idx]]
        return self.dataset[self.indices[idx]]

    def __len__(self):
        return len(self.indices)



import logging
from datetime import datetime
from typing import TYPE_CHECKING

from aim import Image
from aim.pytorch_lightning import AimLogger

if TYPE_CHECKING:
    from nebula.core.engine import Engine


class NebulaLogger(AimLogger):
    def __init__(self, config, engine: "Engine", scenario_start_time, *args, **kwargs):
        self.config = config
        self.engine = engine
        self.scenario_start_time = scenario_start_time
        self.local_step = 0
        self.global_step = 0
        super().__init__(*args, **kwargs)

    def finalize(self, status: str = "") -> None:
        super().finalize(status)
        logging.info(f"Finalizing logger: {status}")

    def get_step(self):
        return int((datetime.now() - datetime.strptime(self.scenario_start_time, "%d/%m/%Y %H:%M:%S")).total_seconds())

    def log_data(self, data, step=None):
        time_start = datetime.now()
        try:
            logging.debug(f"Logging data: {data}")
            super().log_metrics(data)
        except Exception as e:
            logging.exception(f"Error logging statistics data [{data}]: {e}")
        logging.debug(f"Time taken to log data: {datetime.now() - time_start}")

    def log_figure(self, figure, step=None, name=None):
        time_start = datetime.now()
        try:
            logging.debug(f"Logging figure: {name}")
            self.experiment.track(Image(figure), name=name)
        except Exception as e:
            logging.exception(f"Error logging figure: {e}")
        logging.debug(f"Time taken to log figure: {datetime.now() - time_start}")



import matplotlib
import matplotlib.pyplot as plt
from torch import nn
from torchmetrics import MetricCollection

from nebula.core.models.nebulamodel import NebulaModel

matplotlib.use("Agg")
plt.switch_backend("Agg")
import torch
from torchmetrics.classification import (
    MulticlassAccuracy,
    MulticlassConfusionMatrix,
    MulticlassF1Score,
    MulticlassPrecision,
    MulticlassRecall,
)
from torchvision.models import resnet18, resnet34, resnet50

IMAGE_SIZE = 32

BATCH_SIZE = 256 if torch.cuda.is_available() else 64

classifiers = {
    "resnet18": resnet18(),
    "resnet34": resnet34(),
    "resnet50": resnet50(),
}


class CIFAR10ModelResNet(NebulaModel):
    def __init__(
        self,
        input_channels=3,
        num_classes=10,
        learning_rate=1e-3,
        metrics=None,
        confusion_matrix=None,
        seed=None,
        implementation="scratch",
        classifier="resnet9",
    ):
        super().__init__()
        if metrics is None:
            metrics = MetricCollection([
                MulticlassAccuracy(num_classes=num_classes),
                MulticlassPrecision(num_classes=num_classes),
                MulticlassRecall(num_classes=num_classes),
                MulticlassF1Score(num_classes=num_classes),
            ])
        self.train_metrics = metrics.clone(prefix="Train/")
        self.val_metrics = metrics.clone(prefix="Validation/")
        self.test_metrics = metrics.clone(prefix="Test/")

        if confusion_matrix is None:
            self.cm = MulticlassConfusionMatrix(num_classes=num_classes)
        if seed is not None:
            torch.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)

        self.implementation = implementation
        self.classifier = classifier

        self.example_input_array = torch.rand(1, 3, 32, 32)
        self.learning_rate = learning_rate

        self.criterion = torch.nn.CrossEntropyLoss()

        self.model = self._build_model(input_channels, num_classes)

        self.epoch_global_number = {"Train": 0, "Validation": 0, "Test": 0}

    def _build_model(self, input_channels, num_classes):
        if self.implementation == "scratch":
            if self.classifier == "resnet9":
                """
                ResNet9 implementation
                """

                def conv_block(input_channels, num_classes, pool=False):
                    layers = [
                        nn.Conv2d(input_channels, num_classes, kernel_size=3, padding=1),
                        nn.BatchNorm2d(num_classes),
                        nn.ReLU(inplace=True),
                    ]
                    if pool:
                        layers.append(nn.MaxPool2d(2))
                    return nn.Sequential(*layers)

                conv1 = conv_block(input_channels, 64)
                conv2 = conv_block(64, 128, pool=True)
                res1 = nn.Sequential(conv_block(128, 128), conv_block(128, 128))

                conv3 = conv_block(128, 256, pool=True)
                conv4 = conv_block(256, 512, pool=True)
                res2 = nn.Sequential(conv_block(512, 512), conv_block(512, 512))

                classifier = nn.Sequential(nn.MaxPool2d(4), nn.Flatten(), nn.Linear(512, num_classes))

                return nn.ModuleDict({
                    "conv1": conv1,
                    "conv2": conv2,
                    "res1": res1,
                    "conv3": conv3,
                    "conv4": conv4,
                    "res2": res2,
                    "classifier": classifier,
                })

            if self.implementation in classifiers:
                model = classifiers[self.classifier]
                model.fc = torch.nn.Linear(model.fc.in_features, 10)
                return model

            raise NotImplementedError()

        if self.implementation == "timm":
            raise NotImplementedError()

        raise NotImplementedError()

    def forward(self, x):
        if not isinstance(x, torch.Tensor):
            raise TypeError(f"images must be a torch.Tensor, got {type(x)}")

        if self.implementation == "scratch":
            if self.classifier == "resnet9":
                out = self.model["conv1"](x)
                out = self.model["conv2"](out)
                out = self.model["res1"](out) + out
                out = self.model["conv3"](out)
                out = self.model["conv4"](out)
                out = self.model["res2"](out) + out
                out = self.model["classifier"](out)
                return out

            return self.model(x)
        if self.implementation == "timm":
            raise NotImplementedError()

        raise NotImplementedError()

    def configure_optimizers(self):
        if self.implementation == "scratch" and self.classifier == "resnet9":
            params = []
            for key, module in self.model.items():
                params += list(module.parameters())
            optimizer = torch.optim.Adam(params, lr=self.learning_rate, weight_decay=1e-4)
        else:
            optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate, weight_decay=1e-4)
        return optimizer



import torch

from nebula.core.models.nebulamodel import NebulaModel


class CIFAR100ModelCNN(NebulaModel):
    def __init__(
        self,
        input_channels=3,
        num_classes=100,
        learning_rate=1e-3,
        metrics=None,
        confusion_matrix=None,
        seed=None,
    ):
        super().__init__(input_channels, num_classes, learning_rate, metrics, confusion_matrix, seed)

        self.config = {
            "lr": 8.0505e-05,
            "beta1": 0.851436,
            "beta2": 0.999689,
            "amsgrad": True,
        }

        self.example_input_array = torch.rand(1, 3, 32, 32)

        self.criterion = torch.torch.nn.CrossEntropyLoss()

        self.conv1 = torch.nn.Sequential(
            torch.nn.Conv2d(in_channels=3, out_channels=64, kernel_size=3, padding=1),
            torch.nn.BatchNorm2d(64),
            torch.nn.ReLU(inplace=True),
        )
        self.conv2 = torch.nn.Sequential(
            torch.nn.Conv2d(in_channels=64, out_channels=128, kernel_size=3, padding=1),
            torch.nn.BatchNorm2d(128),
            torch.nn.ReLU(inplace=True),
            torch.nn.MaxPool2d(2),
        )

        self.res1 = torch.nn.Sequential(
            torch.nn.Sequential(
                torch.nn.Conv2d(in_channels=128, out_channels=128, kernel_size=3, padding=1),
                torch.nn.BatchNorm2d(128),
                torch.nn.ReLU(inplace=True),
            ),
            torch.nn.Sequential(
                torch.nn.Conv2d(in_channels=128, out_channels=128, kernel_size=3, padding=1),
                torch.nn.BatchNorm2d(128),
                torch.nn.ReLU(inplace=True),
            ),
        )

        self.conv3 = torch.nn.Sequential(
            torch.nn.Conv2d(in_channels=128, out_channels=256, kernel_size=3, padding=1),
            torch.nn.BatchNorm2d(256),
            torch.nn.ReLU(inplace=True),
            torch.nn.MaxPool2d(2),
        )
        self.conv4 = torch.nn.Sequential(
            torch.nn.Conv2d(in_channels=256, out_channels=512, kernel_size=3, padding=1),
            torch.nn.BatchNorm2d(512),
            torch.nn.ReLU(inplace=True),
            torch.nn.MaxPool2d(2),
        )

        self.res2 = torch.nn.Sequential(
            torch.nn.Sequential(
                torch.nn.Conv2d(in_channels=512, out_channels=512, kernel_size=3, padding=1),
                torch.nn.BatchNorm2d(512),
                torch.nn.ReLU(inplace=True),
            ),
            torch.nn.Sequential(
                torch.nn.Conv2d(in_channels=512, out_channels=512, kernel_size=3, padding=1),
                torch.nn.BatchNorm2d(512),
                torch.nn.ReLU(inplace=True),
            ),
        )

        self.classifier = torch.nn.Sequential(
            torch.nn.MaxPool2d(4),
            torch.nn.Flatten(),
            torch.nn.Linear(512, self.num_classes),
        )

    def forward(self, x):
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.res1(x) + x
        x = self.conv3(x)
        x = self.conv4(x)
        x = self.res2(x) + x
        x = self.classifier(x)
        return x

    def configure_optimizers(self):
        return torch.optim.Adam(
            self.parameters(),
            lr=self.config["lr"],
            betas=(self.config["beta1"], self.config["beta2"]),
            amsgrad=self.config["amsgrad"],
        )



import asyncio
import logging
import time
from typing import TYPE_CHECKING

from nebula.addons.functions import print_msg_box
from nebula.core.pb import nebula_pb2

if TYPE_CHECKING:
    from nebula.core.network.communications import CommunicationsManager


class Health:
    def __init__(self, addr, config, cm: "CommunicationsManager"):
        print_msg_box(msg="Starting health module...", indent=2, title="Health module")
        self.addr = addr
        self.config = config
        self.cm = cm
        self.period = self.config.participant["health_args"]["health_interval"]
        self.alive_interval = self.config.participant["health_args"]["send_alive_interval"]
        self.check_alive_interval = self.config.participant["health_args"]["check_alive_interval"]
        self.timeout = self.config.participant["health_args"]["alive_timeout"]

    async def start(self):
        asyncio.create_task(self.run_send_alive())
        asyncio.create_task(self.run_check_alive())

    async def run_send_alive(self):
        await asyncio.sleep(self.config.participant["health_args"]["grace_time_health"])
        # Set all connections to active at the beginning of the health module
        for conn in self.cm.connections.values():
            conn.set_active(True)
        while True:
            if len(self.cm.connections) > 0:
                message = self.cm.mm.generate_control_message(
                    nebula_pb2.ControlMessage.Action.ALIVE, log="Alive message"
                )
                current_connections = list(self.cm.connections.values())
                for conn in current_connections:
                    if conn.get_direct():
                        try:
                            logging.info(f"🕒  Sending alive message to {conn.get_addr()}...")
                            await conn.send(data=message)
                        except Exception as e:
                            logging.exception(f"❗️  Cannot send alive message to {conn.get_addr()}. Error: {e!s}")
                    await asyncio.sleep(self.alive_interval)
            await asyncio.sleep(self.period)

    async def run_check_alive(self):
        await asyncio.sleep(self.config.participant["health_args"]["grace_time_health"] + self.check_alive_interval)
        while True:
            if len(self.cm.connections) > 0:
                current_connections = list(self.cm.connections.values())
                for conn in current_connections:
                    if conn.get_direct():
                        if time.time() - conn.get_last_active() > self.timeout:
                            logging.error(f"⬅️ 🕒  Heartbeat timeout for {conn.get_addr()}...")
                            await self.cm.disconnect(conn.get_addr(), mutual_disconnection=False)
            await asyncio.sleep(self.check_alive_interval)

    async def alive(self, source):
        current_time = time.time()
        if source not in self.cm.connections:
            logging.error(f"❗️  Connection {source} not found in connections...")
            return
        conn = self.cm.connections[source]
        if conn.get_last_active() < current_time:
            logging.debug(f"🕒  Updating last active time for {source}")
            conn.set_active(True)



import base64
import json
import logging
import os

import requests


class Mender:
    def __init__(self):
        self.server = os.environ.get("MENDER_SERVER")
        self.user = os.environ.get("MENDER_USER")
        self.password = os.environ.get("MENDER_PASSWORD")
        self.token = os.environ.get("MENDER_TOKEN")
        logging.info(f"Mender server: {self.server}")

    def get_token(self):
        return self.token

    def renew_token(self):
        string = self.user + ":" + self.password
        base64string = base64.b64encode(string.encode("utf-8"))
        headers = {
            "Accept": "application/json",
            "Authorization": "Basic {}".format(str(base64string, "utf-8")),
            "Content-Type": "application/json",
        }

        r = requests.post(f"{self.server}/api/management/v1/useradm/auth/login", headers=headers)
        self.token = r.text

    @staticmethod
    def generate_artifact(type_artifact, artifact_name, device_type, file_path):
        os.system(
            f"mender-artifact write module-image -T {type_artifact} -n {artifact_name} -t {device_type} -o {artifact_name}.mender -f {file_path}"
        )

    def get_artifacts(self):
        headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {self.get_token()}",
        }

        r = requests.get(f"{self.server}/api/management/v1/deployments/artifacts", headers=headers)

        logging.info(json.dumps(r.json(), indent=2))

    def upload_artifact(self, artifact_path, description):
        headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {self.get_token()}",
        }

        multipart_form_data = {
            "description": (None, f"{description}"),
            "artifact": (None, open(f"{artifact_path}", "rb")),
        }

        r = requests.post(
            f"{self.server}/api/management/v1/deployments/artifacts",
            files=multipart_form_data,
            headers=headers,
        )
        logging.info(r.text)

    def deploy_artifact_device(self, artifact_name, device):
        headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {self.get_token()}",
            "Content-Type": "application/json",
        }

        json_req = {
            "artifact_name": f"{artifact_name}",
            "devices": [f"{device}"],
            "name": f"{device}",
            "all_devices": False,
        }

        r = requests.post(
            f"{self.server}/api/management/v1/deployments/deployments",
            json=json_req,
            headers=headers,
        )

        logging.info(r.text)

    def deploy_artifact_list(self, artifact_name, devices: list[str]):
        headers = {
            "Accept": "application/json",
            "Authorization": "Basic {}".format((self.user + ":" + self.password).encode("base64")),
            "Content-Type": "application/form-data",
        }

        json_req = {
            "artifact_name": f"{artifact_name}",
            "devices": devices,
            "name": "Example_deployment",
        }

        r = requests.post(
            f"{self.server}/api/management/v1/deployments/deployments",
            json=json_req,
            headers=headers,
        )

        logging.info(r.text)

    def get_info_deployment(self, deployment_id):
        headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {self.get_token()}",
        }

        r = requests.get(
            f"{self.server}/api/management/v1/deployments/deployments/{deployment_id}",
            headers=headers,
        )

        logging.info(json.dumps(r.json(), indent=2))

    def get_my_info(self):
        headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {self.get_token()}",
        }

        r = requests.get(f"{self.server}/api/management/v1/useradm/users/me", headers=headers)

        logging.info(json.dumps(r.json(), indent=2))

    def get_devices(self):
        headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {self.get_token()}",
        }

        r = requests.get(
            f"{self.server}/api/management/v2/devauth/devices",
            params={"per_page": "300"},
            headers=headers,
        )

        logging.info(json.dumps(r.json(), indent=2))

    def get_devices_by_group(self, group):
        headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {self.get_token()}",
        }

        json_req = {
            "page": 1,
            "per_page": 300,
            "filters": [
                {
                    "scope": "system",
                    "attribute": "group",
                    "type": "$eq",
                    "value": group,
                },
                {
                    "scope": "identity",
                    "attribute": "status",
                    "type": "$eq",
                    "value": "accepted",
                },
            ],
            "sort": [],
            "attributes": [
                {"scope": "identity", "attribute": "status"},
                {"scope": "inventory", "attribute": "artifact_name"},
                {"scope": "inventory", "attribute": "device_type"},
                {"scope": "inventory", "attribute": "rootfs-image.version"},
                {"scope": "monitor", "attribute": "alerts"},
                {"scope": "system", "attribute": "created_ts"},
                {"scope": "system", "attribute": "updated_ts"},
                {"scope": "tags", "attribute": "name"},
                {"scope": "identity", "attribute": "name"},
            ],
        }

        r = requests.post(
            f"{self.server}/api/management/v2/inventory/filters/search",
            json=json_req,
            headers=headers,
        )

        logging.info(json.dumps(r.json(), indent=2))

        # json to file
        with open(f"devices_{group}.json", "w") as outfile:
            json.dump(r.json(), outfile, indent=2)

    def get_info_device(self, device_id):
        headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {self.get_token()}",
        }

        r = requests.get(
            f"{self.server}/api/management/v1/inventory/devices/{device_id}",
            headers=headers,
        )

        logging.info(json.dumps(r.json(), indent=2))

    def get_connected_device(self, device_id):
        headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {self.get_token()}",
        }

        r = requests.get(
            f"{self.server}/api/management/v1/deviceconnect/devices/{device_id}",
            headers=headers,
        )

        logging.info(json.dumps(r.json(), indent=2))



import os
import sys
import urllib.request

import numpy as np
import torch
from torchvision.datasets import MNIST

from nebula.core.datasets.nebuladataset import NebulaDataset


class WADI(MNIST):
    def __init__(self, partition_id, partitions_number, root_dir, train=True):
        super(MNIST, self).__init__(root_dir, transform=None, target_transform=None)
        self.partition_id = partition_id
        self.partitions_number = partitions_number
        self.download_link = "XXXX"
        self.files = ["X_train.npy", "y_train.npy", "X_test.npy", "y_test.npy"]
        self.train = train
        self.root = root_dir

        if (
            not os.path.exists(f"{self.root}/WADI/X_train.npy")
            or not os.path.exists(f"{self.root}/WADI/y_train.npy")
            or not os.path.exists(f"{self.root}/WADI/X_test.npy")
            or not os.path.exists(f"{self.root}/WADI/y_test.npy")
        ):
            self.dataset_download()

        if self.train:
            data_file = self.training_file
            self.data, self.targets = (
                torch.from_numpy(np.load(f"{self.root}/WADI/X_train.npy")),
                torch.from_numpy(np.load(f"{self.root}/WADI/y_train.npy")),
            )
            self.data = self.data.to(torch.float32)
            self.targets = self.targets.to(torch.float32)
        else:
            data_file = self.test_file
            self.data, self.targets = (
                torch.from_numpy(np.load(f"{self.root}/WADI/X_test.npy")),
                torch.from_numpy(np.load(f"{self.root}/WADI/y_test.npy")),
            )
            self.data = self.data.to(torch.float32)
            self.targets = self.targets.to(torch.float32)

    def __getitem__(self, index):
        img, target = self.data[index], int(self.targets[index])
        return img, target

    def dataset_download(self):
        paths = [f"{self.root}/WADI/"]
        for path in paths:
            if not os.path.exists(path):
                os.makedirs(path)
        for file in self.files:
            urllib.request.urlretrieve(
                os.path.join(f"{self.download_link}", file),
                os.path.join(f"{self.root}/WADI/", file),
            )


class WADIDataModule(NebulaDataset):
    def __init__(
        self,
        num_classes=10,
        partition_id=0,
        partitions_number=1,
        batch_size=32,
        num_workers=4,
        iid=True,
        partition="dirichlet",
        partition_parameter=0.5,
        seed=42,
        config=None,
    ):
        super().__init__(
            num_classes=num_classes,
            partition_id=partition_id,
            partitions_number=partitions_number,
            batch_size=batch_size,
            num_workers=num_workers,
            iid=iid,
            partition=partition,
            partition_parameter=partition_parameter,
            seed=seed,
            config=config,
        )

    def initialize_dataset(self):
        # Load wadi train dataset
        if self.train_set is None:
            self.train_set = self.load_wadi_dataset(train=True)
        if self.test_set is None:
            self.test_set = self.load_wadi_dataset(train=False)

        # All nodes have the same test set (indices are the same for all nodes)
        self.test_indices_map = list(range(len(self.test_set)))

        # Depending on the iid flag, generate a non-iid or iid map of the train set
        if self.iid:
            self.train_indices_map = self.generate_iid_map(self.train_set, self.partition, self.partition_parameter)
            self.local_test_indices_map = self.generate_iid_map(self.test_set, self.partition, self.partition_parameter)
        else:
            self.train_indices_map = self.generate_non_iid_map(self.train_set, self.partition, self.partition_parameter)
            self.local_test_indices_map = self.generate_non_iid_map(
                self.test_set, self.partition, self.partition_parameter
            )

        print(f"Length of train indices map: {len(self.train_indices_map)}")
        print(f"Lenght of test indices map: {len(self.test_indices_map)}")
        print(f"Lenght of test indices map (global): {len(self.test_indices_map)}")
        print(f"Length of test indices map (local): {len(self.local_test_indices_map)}")

    def load_wadi_dataset(self, train=True):
        if train:
            return WADI(
                partition_id=self.partition_id,
                partitions_number=self.partitions_number,
                root_dir=f"{sys.path[0]}/data",
                train=True,
            )
        else:
            return WADI(
                partition_id=self.partition_id,
                partitions_number=self.partitions_number,
                root_dir=f"{sys.path[0]}/data",
                train=False,
            )

    def generate_non_iid_map(self, dataset, partition="dirichlet", partition_parameter=0.5):
        if partition == "dirichlet":
            partitions_map = self.dirichlet_partition(dataset, alpha=partition_parameter)
        elif partition == "percent":
            partitions_map = self.percentage_partition(dataset, percentage=partition_parameter)
        else:
            raise ValueError(f"Partition {partition} is not supported for Non-IID map")

        if self.partition_id == 0:
            self.plot_data_distribution(dataset, partitions_map)
            self.plot_all_data_distribution(dataset, partitions_map)

        return partitions_map[self.partition_id]

    def generate_iid_map(self, dataset, partition="balancediid", partition_parameter=2):
        if partition == "balancediid":
            partitions_map = self.balanced_iid_partition(dataset)
        elif partition == "unbalancediid":
            partitions_map = self.unbalanced_iid_partition(dataset, imbalance_factor=partition_parameter)
        else:
            raise ValueError(f"Partition {partition} is not supported for IID map")

        if self.partition_id == 0:
            self.plot_data_distribution(dataset, partitions_map)
            self.plot_all_data_distribution(dataset, partitions_map)

        return partitions_map[self.partition_id]



import asyncio
import copy
import gc
import gzip
import hashlib
import io
import logging
import os
import pickle
import traceback
from collections import OrderedDict

import torch
from lightning import Trainer
from lightning.pytorch.callbacks import ModelSummary, ProgressBar
from lightning.pytorch.loggers import CSVLogger
from torch.nn import functional as F

from nebula.core.utils.deterministic import enable_deterministic
from nebula.core.utils.nebulalogger_tensorboard import NebulaTensorBoardLogger

try:
    from nebula.core.utils.nebulalogger import NebulaLogger
except:
    pass
from nebula.config.config import TRAINING_LOGGER

logging_training = logging.getLogger(TRAINING_LOGGER)


class NebulaProgressBar(ProgressBar):
    """Nebula progress bar for training.
    Logs the percentage of completion of the training process using logging.
    """

    def __init__(self, log_every_n_steps=100):
        super().__init__()
        self.enable = True
        self.log_every_n_steps = log_every_n_steps

    def enable(self):
        """Enable progress bar logging."""
        self.enable = True

    def disable(self):
        """Disable the progress bar logging."""
        self.enable = False

    def on_train_epoch_start(self, trainer, pl_module):
        """Called when the training epoch starts."""
        super().on_train_epoch_start(trainer, pl_module)
        if self.enable:
            logging_training.info(f"Starting Epoch {trainer.current_epoch}")

    def on_train_batch_end(self, trainer, pl_module, outputs, batch, batch_idx):
        """Called at the end of each training batch."""
        super().on_train_batch_end(trainer, pl_module, outputs, batch, batch_idx)
        if self.enable:
            if (batch_idx + 1) % self.log_every_n_steps == 0 or (batch_idx + 1) == self.total_train_batches:
                # Calculate percentage complete for the current epoch
                percent = ((batch_idx + 1) / self.total_train_batches) * 100  # +1 to count current batch
                logging_training.info(f"Epoch {trainer.current_epoch} - {percent:.01f}% complete")

    def on_train_epoch_end(self, trainer, pl_module):
        """Called at the end of the training epoch."""
        super().on_train_epoch_end(trainer, pl_module)
        if self.enable:
            logging_training.info(f"Epoch {trainer.current_epoch} finished")

    def on_validation_epoch_start(self, trainer, pl_module):
        super().on_validation_epoch_start(trainer, pl_module)
        if self.enable:
            logging_training.info(f"Starting validation for Epoch {trainer.current_epoch}")

    def on_validation_epoch_end(self, trainer, pl_module):
        super().on_validation_epoch_end(trainer, pl_module)
        if self.enable:
            logging_training.info(f"Validation for Epoch {trainer.current_epoch} finished")

    def on_test_batch_start(self, trainer, pl_module, batch, batch_idx, dataloader_idx):
        super().on_test_batch_start(trainer, pl_module, batch, batch_idx, dataloader_idx)
        if not self.has_dataloader_changed(dataloader_idx):
            return

    def on_test_batch_end(self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx):
        """Called at the end of each test batch."""
        super().on_test_batch_end(trainer, pl_module, outputs, batch, batch_idx, dataloader_idx)
        if self.enable:
            total_batches = self.total_test_batches_current_dataloader
            if total_batches == 0:
                logging_training.warning(
                    f"Total test batches is 0 for dataloader {dataloader_idx}, cannot compute progress."
                )
                return

            if (batch_idx + 1) % self.log_every_n_steps == 0 or (batch_idx + 1) == total_batches:
                percent = ((batch_idx + 1) / total_batches) * 100  # +1 to count the current batch
                logging_training.info(
                    f"Test Epoch {trainer.current_epoch}, Dataloader {dataloader_idx} - {percent:.01f}% complete"
                )

    def on_test_epoch_start(self, trainer, pl_module):
        super().on_test_epoch_start(trainer, pl_module)
        if self.enable:
            logging_training.info(f"Starting testing for Epoch {trainer.current_epoch}")

    def on_test_epoch_end(self, trainer, pl_module):
        super().on_test_epoch_end(trainer, pl_module)
        if self.enable:
            logging_training.info(f"Testing for Epoch {trainer.current_epoch} finished")


class ParameterSerializeError(Exception):
    """Custom exception for errors setting model parameters."""


class ParameterDeserializeError(Exception):
    """Custom exception for errors setting model parameters."""


class ParameterSettingError(Exception):
    """Custom exception for errors setting model parameters."""


class Lightning:
    DEFAULT_MODEL_WEIGHT = 1
    BYPASS_MODEL_WEIGHT = 0

    def __init__(self, model, data, config=None):
        # self.model = torch.compile(model, mode="reduce-overhead")
        self.model = model
        self.data = data
        self.config = config
        self._trainer = None
        self.epochs = 1
        self.round = 0
        self.experiment_name = self.config.participant["scenario_args"]["name"]
        self.idx = self.config.participant["device_args"]["idx"]
        self.log_dir = os.path.join(self.config.participant["tracking_args"]["log_dir"], self.experiment_name)
        self._logger = None
        self.create_logger()
        enable_deterministic(self.config)

    @property
    def logger(self):
        return self._logger

    def get_round(self):
        return self.round

    def set_model(self, model):
        self.model = model

    def set_data(self, data):
        self.data = data

    def create_logger(self):
        if self.config.participant["tracking_args"]["local_tracking"] == "csv":
            nebulalogger = CSVLogger(f"{self.log_dir}", name="metrics", version=f"participant_{self.idx}")
        elif self.config.participant["tracking_args"]["local_tracking"] == "basic":
            logger_config = None
            if self._logger is not None:
                logger_config = self._logger.get_logger_config()
            nebulalogger = NebulaTensorBoardLogger(
                self.config.participant["scenario_args"]["start_time"],
                f"{self.log_dir}",
                name="metrics",
                version=f"participant_{self.idx}",
                log_graph=False,
            )
            # Restore logger configuration
            nebulalogger.set_logger_config(logger_config)
        elif self.config.participant["tracking_args"]["local_tracking"] == "advanced":
            nebulalogger = NebulaLogger(
                config=self.config,
                engine=self,
                scenario_start_time=self.config.participant["scenario_args"]["start_time"],
                repo=f"{self.config.participant['tracking_args']['log_dir']}",
                experiment=self.experiment_name,
                run_name=f"participant_{self.idx}",
                train_metric_prefix="train_",
                test_metric_prefix="test_",
                val_metric_prefix="val_",
                log_system_params=False,
            )
            # nebulalogger_aim = NebulaLogger(config=self.config, engine=self, scenario_start_time=self.config.participant["scenario_args"]["start_time"], repo=f"aim://nebula-frontend:8085",
            #                                     experiment=self.experiment_name, run_name=f"participant_{self.idx}",
            #                                     train_metric_prefix='train_', test_metric_prefix='test_', val_metric_prefix='val_', log_system_params=False)
            self.config.participant["tracking_args"]["run_hash"] = nebulalogger.experiment.hash
        else:
            nebulalogger = None

        self._logger = nebulalogger

    def create_trainer(self):
        # Create a new trainer and logger for each round
        self.create_logger()
        num_gpus = torch.cuda.device_count()
        if self.config.participant["device_args"]["accelerator"] == "gpu" and num_gpus > 0:
            gpu_index = self.config.participant["device_args"]["idx"] % num_gpus
            logging_training.info(f"Creating trainer with accelerator GPU ({gpu_index})")
            self._trainer = Trainer(
                callbacks=[ModelSummary(max_depth=1), NebulaProgressBar()],
                max_epochs=self.epochs,
                accelerator=self.config.participant["device_args"]["accelerator"],
                devices=[gpu_index],
                logger=self._logger,
                enable_checkpointing=False,
                enable_model_summary=False,
                # deterministic=True
            )
        else:
            logging_training.info("Creating trainer with accelerator CPU")
            self._trainer = Trainer(
                callbacks=[ModelSummary(max_depth=1), NebulaProgressBar()],
                max_epochs=self.epochs,
                accelerator=self.config.participant["device_args"]["accelerator"],
                devices="auto",
                logger=self._logger,
                enable_checkpointing=False,
                enable_model_summary=False,
                # deterministic=True
            )
        logging_training.info(f"Trainer strategy: {self._trainer.strategy}")

    def validate_neighbour_model(self, neighbour_model_param):
        avg_loss = 0
        running_loss = 0
        bootstrap_dataloader = self.data.bootstrap_dataloader()
        num_samples = 0
        neighbour_model = copy.deepcopy(self.model)
        neighbour_model.load_state_dict(neighbour_model_param)

        # enable evaluation mode, prevent memory leaks.
        # no need to switch back to training since model is not further used.
        if torch.cuda.is_available():
            neighbour_model = neighbour_model.to("cuda")
        neighbour_model.eval()

        # bootstrap_dataloader = bootstrap_dataloader.to('cuda')
        with torch.no_grad():
            for inputs, labels in bootstrap_dataloader:
                if torch.cuda.is_available():
                    inputs = inputs.to("cuda")
                    labels = labels.to("cuda")
                outputs = neighbour_model(inputs)
                loss = F.cross_entropy(outputs, labels)
                running_loss += loss.item()
                num_samples += inputs.size(0)

        avg_loss = running_loss / len(bootstrap_dataloader)
        logging_training.info(f"Computed neighbor loss over {num_samples} data samples")
        return avg_loss

    def get_hash_model(self):
        """
        Returns:
            str: SHA256 hash of model parameters
        """
        return hashlib.sha256(self.serialize_model(self.model)).hexdigest()

    def set_epochs(self, epochs):
        self.epochs = epochs

    def serialize_model(self, model):
        # From https://pytorch.org/docs/stable/notes/serialization.html
        try:
            buffer = io.BytesIO()
            with gzip.GzipFile(fileobj=buffer, mode="wb") as f:
                torch.save(model, f, pickle_protocol=pickle.HIGHEST_PROTOCOL)
            serialized_data = buffer.getvalue()
            buffer.close()
            del buffer
            return serialized_data
        except Exception as e:
            raise ParameterSerializeError("Error serializing model") from e

    def deserialize_model(self, data):
        # From https://pytorch.org/docs/stable/notes/serialization.html
        try:
            buffer = io.BytesIO(data)
            with gzip.GzipFile(fileobj=buffer, mode="rb") as f:
                params_dict = torch.load(f)
            buffer.close()
            del buffer
            return OrderedDict(params_dict)
        except Exception as e:
            raise ParameterDeserializeError("Error decoding parameters") from e

    def set_model_parameters(self, params, initialize=False):
        try:
            self.model.load_state_dict(params)
        except Exception as e:
            raise ParameterSettingError("Error setting parameters") from e

    def get_model_parameters(self, bytes=False, initialize=False):
        if bytes:
            return self.serialize_model(self.model.state_dict())
        return self.model.state_dict()

    async def train(self):
        try:
            self.create_trainer()
            logging.info(f"{'=' * 10} [Training] Started (check training logs for progress) {'=' * 10}")
            await asyncio.to_thread(self._train_sync)
            logging.info(f"{'=' * 10} [Training] Finished (check training logs for progress) {'=' * 10}")
        except Exception as e:
            logging_training.error(f"Error training model: {e}")
            logging_training.error(traceback.format_exc())

    def _train_sync(self):
        try:
            self._trainer.fit(self.model, self.data)
        except Exception as e:
            logging_training.error(f"Error in _train_sync: {e}")
            tb = traceback.format_exc()
            logging_training.error(f"Traceback: {tb}")
            # If "raise", the exception will be managed by the main thread

    async def test(self):
        try:
            self.create_trainer()
            logging.info(f"{'=' * 10} [Testing] Started (check training logs for progress) {'=' * 10}")
            await asyncio.to_thread(self._test_sync)
            logging.info(f"{'=' * 10} [Testing] Finished (check training logs for progress) {'=' * 10}")
        except Exception as e:
            logging_training.error(f"Error testing model: {e}")
            logging_training.error(traceback.format_exc())

    def _test_sync(self):
        try:
            self._trainer.test(self.model, self.data, verbose=True)
        except Exception as e:
            logging_training.error(f"Error in _test_sync: {e}")
            tb = traceback.format_exc()
            logging_training.error(f"Traceback: {tb}")
            # If "raise", the exception will be managed by the main thread

    def cleanup(self):
        if self._trainer is not None:
            self._trainer._teardown()
            del self._trainer
        if self.data is not None:
            self.data.teardown()
        gc.collect()
        torch.cuda.empty_cache()

    def get_model_weight(self):
        weight = self.data.model_weight
        if weight is None:
            raise ValueError("Model weight not set. Please call setup('fit') before requesting model weight.")
        return weight

    def on_round_start(self):
        self.data.setup()
        self._logger.log_data({"Round": self.round})
        # self.reporter.enqueue_data("Round", self.round)

    def on_round_end(self):
        self._logger.global_step = self._logger.global_step + self._logger.local_step
        self._logger.local_step = 0
        self.round += 1
        self.model.on_round_end()
        logging.info("Flushing memory cache at the end of round...")
        self.cleanup()

    def on_learning_cycle_end(self):
        self._logger.log_data({"Round": self.round})
        # self.reporter.enqueue_data("Round", self.round)



import time
from collections import OrderedDict
from collections.abc import Mapping

import requests
import torch
from eth_account import Account
from retry import retry
from tabulate import tabulate
from web3 import Web3
from web3.middleware import construct_sign_and_send_raw_middleware, geth_poa_middleware

from nebula.core.aggregation.aggregator import Aggregator
from nebula.core.utils.helper import (
    cosine_metric,
    euclidean_metric,
    jaccard_metric,
    manhattan_metric,
    minkowski_metric,
    pearson_correlation_metric,
)


def cossim_euclidean(model1, model2, similarity):
    return cosine_metric(model1, model2, similarity=similarity) * euclidean_metric(
        model1, model2, similarity=similarity
    )


class BlockchainReputation(Aggregator):
    """
    # BAT-SandrinHunkeler (BlockchainReputation)
    Weighted FedAvg by using relative reputation of each model's trainer
    Returns: aggregated model
    """

    ALGORITHM_MAP = {
        "Cossim": cosine_metric,
        "Pearson": pearson_correlation_metric,
        "Euclidean": euclidean_metric,
        "Minkowski": minkowski_metric,
        "Manhattan": manhattan_metric,
        "Jaccard": jaccard_metric,
        "CossimEuclid": cossim_euclidean,
    }

    def __init__(self, similarity_metric: str = "CossimEuclid", config=None, **kwargs):
        # initialize parent class
        super().__init__(config, **kwargs)

        self.config = config

        # extract local NEBULA name
        self.node_name = self.config.participant["network_args"]["addr"]

        # initialize BlockchainHandler for interacting with oracle and non-validator node
        self.__blockchain = BlockchainHandler(self.node_name)

        # check if node is malicious for debugging
        self.__malicious = self.config.participant["device_args"]["malicious"]

        self.__opinion_algo = BlockchainReputation.ALGORITHM_MAP.get(similarity_metric)
        self.__similarity_metric = similarity_metric

    def run_aggregation(self, model_buffer: OrderedDict[str, OrderedDict[torch.Tensor, int]]) -> torch.Tensor:
        print_with_frame("BLOCKCHAIN AGGREGATION: START")

        # track aggregation time for experiments
        start = time.time_ns()

        # verify the registration process during initialization of the BlockchainHandler
        self.__blockchain.verify_registration()

        # verify if ether balance is still sufficient for aggregating, request more otherwise
        self.__blockchain.verify_balance()

        # create dict<sender, model>
        current_models = {sender: model for sender, (model, weight) in model_buffer.items()}

        print(f"Node: {self.node_name}", flush=True)
        print(f"self.__malicious: {self.__malicious}", flush=True)

        # extract local model from models to aggregate
        local_model = model_buffer[self.node_name][0]

        # compute similarity between local model and all buffered models
        metric_values = {
            sender: max(
                min(
                    round(
                        self.__opinion_algo(local_model, current_models[sender], similarity=True),
                        5,
                    ),
                    1,
                ),
                0,
            )
            for sender in current_models
            if sender != self.node_name
        }

        # log similarity metric values
        print_table(
            "SIMILARITY METRIC",
            list(metric_values.items()),
            ["neighbour Node", f"{self.__similarity_metric} Similarity"],
        )

        # increase resolution of metric in upper half of interval
        opinion_values = {sender: round(metric**3 * 100) for sender, metric in metric_values.items()}

        # DEBUG
        if int(self.node_name[-7]) <= 1 and self.__blockchain.round >= 5:
            opinion_values = {sender: int(torch.randint(0, 101, (1,))[0]) for sender, metric in metric_values.items()}

        # push local opinions to reputation system
        self.__blockchain.push_opinions(opinion_values)

        # log pushed opinion values
        print_table(
            "REPORT LOCAL OPINION",
            list(opinion_values.items()),
            ["Node", f"Transformed {self.__similarity_metric} Similarity"],
        )

        # request global reputation values from reputation system
        reputation_values = self.__blockchain.get_reputations([sender for sender in current_models])

        # log received global reputations
        print_table(
            "GLOBAL REPUTATION",
            list(reputation_values.items()),
            ["Node", "Global Reputation"],
        )

        # normalize all reputation values to sum() == 1
        sum_reputations = sum(reputation_values.values())
        if sum_reputations > 0:
            normalized_reputation_values = {
                name: round(reputation_values[name] / sum_reputations, 3) for name in reputation_values
            }
        else:
            normalized_reputation_values = reputation_values

        # log normalized aggregation weights
        print_table(
            "AGGREGATION WEIGHTS",
            list(normalized_reputation_values.items()),
            ["Node", "Aggregation Weight"],
        )

        # initialize empty model
        final_model = {layer: torch.zeros_like(param).float() for layer, param in local_model.items()}

        # cover rare case where no models were added or reputation is 0 to return the local model
        if sum_reputations > 0:
            for sender in normalized_reputation_values.keys():
                for layer in final_model:
                    final_model[layer] += current_models[sender][layer].float() * normalized_reputation_values[sender]

        # otherwise, skip aggregation
        else:
            final_model = local_model

        # report used gas to oracle and log cumulative gas used
        print_table(
            "TOTAL GAS USED",
            self.__blockchain.report_gas_oracle(),
            ["Node", "Cumulative Gas used"],
        )
        self.__blockchain.report_time_oracle(start)

        print_with_frame("BLOCKCHAIN AGGREGATION: FINISHED")

        # return newly aggregated model
        return final_model


def print_table(title: str, values: list[tuple | list], headers: list[str]) -> None:
    """
    Prints a title, all values ordered in a table, with the headers as column titles.
    Args:
        title: Title of the table
        values: Rows of table
        headers: Column headers of table

    Returns: None, prints output

    """
    print(f"\n{'-' * 25} {title.upper()} {'-' * 25}", flush=True)
    print(tabulate(sorted(values), headers=headers, tablefmt="grid"), flush=True)


def print_with_frame(message) -> None:
    """
    Prints a large frame with a title inside
    Args:
        message: Title to put into the frame

    Returns: None

    """
    message_length = len(message)
    print(f"{' ' * 20}+{'-' * (message_length + 2)}+", flush=True)
    print(f"{'*' * 20}| {message.upper()} |{'*' * 20}", flush=True)
    print(f"{' ' * 20}+{'-' * (message_length + 2)}+", flush=True)


class BlockchainHandler:
    """
    Handles interaction with Oracle and Non-Validator Node of Blockchain Network
    """

    # static ip address of non-validator node with RPC-API
    __rpc_url = "http://172.25.0.104:8545"

    # static ip address of oracle with REST-API
    __oracle_url = "http://172.25.0.105:8081"

    # default REST header for interacting with oracle
    __rest_header = {"Content-type": "application/json", "Accept": "application/json"}

    def __init__(self, home_address):
        print_with_frame("BLOCKCHAIN INITIALIZATION: START")

        # local NEBULA name, needed for reputation system
        self.__home_ip = home_address

        # randomly generated private key, needed to sign transaction
        self.__private_key = ""

        # public wallet address generated from the private key
        self.__acc_address = ""

        # variables for experiment, not required for aggregation
        self.__gas_used = 0
        self.__gas_price = 27.3
        self.round = 1

        # generate randomized primary key
        self.__acc = self.__create_account()

        # configure web3 objects for using Proof-of-Authority
        self.__web3 = self.__initialize_web3()

        # call Oracle to sense if blockchain is ready
        print(f"{'-' * 25} CONNECT TO ORACLE {'-' * 25}", flush=True)
        self.__wait_for_blockchain()

        # request ETH funds for creating transactions, paying gas
        self.__request_funds_from_oracle()

        # check if funds were assigned by checking directly with blockchain
        self.verify_balance()

        # request contract address and header from Oracle
        self.__contract_obj = self.__get_contract_from_oracle()

        # register public wallet address at reputation system
        print(f"{'-' * 25} CONNECT TO REPUTATION SYSTEM {'-' * 25}", flush=True)
        self.__register()
        print("BLOCKCHAIN: Registered to reputation system", flush=True)

        # check if registration was successful
        self.verify_registration()
        print("BLOCKCHAIN: Verified registration", flush=True)

        print_with_frame("BLOCKCHAIN INITIALIZATION: FINISHED")

    @classmethod
    @property
    def oracle_url(cls) -> str:
        return cls.__oracle_url

    @classmethod
    @property
    def rest_header(cls) -> Mapping[str, str]:
        return cls.__rest_header

    def __create_account(self):
        """
        Generates randomized primary key and derives public account from it
        Returns: None

        """
        print(f"{'-' * 25} REGISTER WORKING NODE {'-' * 25}", flush=True)

        # generate random private key, address, public address
        acc = Account.create()

        # initialize web3 utility object
        web3 = Web3()

        # convert private key to hex, used in raw transactions
        self.__private_key = web3.to_hex(acc.key)

        # convert address type, used in raw transactions
        self.__acc_address = web3.to_checksum_address(acc.address)

        print(f"WORKER NODE: Registered account: {self.__home_ip}", flush=True)
        print(f"WORKER NODE: Account address: {self.__acc_address}", flush=True)

        # return generated account
        return acc

    def __initialize_web3(self):
        """
        Initializes Web3 object and configures it for PoA protocol
        Returns: Web3 object

        """

        # initialize Web3 object with ip of non-validator node
        web3 = Web3(Web3.HTTPProvider(self.__rpc_url, request_kwargs={"timeout": 20}))  # 10

        # inject Proof-of-Authority settings to object
        web3.middleware_onion.inject(geth_poa_middleware, layer=0)

        # automatically sign transactions if available for execution
        web3.middleware_onion.add(construct_sign_and_send_raw_middleware(self.__acc))

        # inject local account as default
        web3.eth.default_account = self.__acc_address

        # return initialized object for executing transaction
        return web3

    @retry((Exception, requests.exceptions.HTTPError), tries=20, delay=4)
    def __wait_for_blockchain(self) -> None:
        """
        Request state of blockchain from Oracle by periodic calls and sleep
        Returns: None

        """

        # check with oracle if blockchain is ready for requests
        response = requests.get(
            url=f"{self.__oracle_url}/status",
            headers=self.__rest_header,
            timeout=20,  # 10
        )

        # raise Exception if status is not successful
        response.raise_for_status()

        return print("ORACLE: Blockchain is ready", flush=True)

    @retry((Exception, requests.exceptions.HTTPError), tries=3, delay=4)
    def __request_funds_from_oracle(self) -> None:
        """
        Requests funds from Oracle by sending public address
        Returns: None

        """

        # call oracle's faucet by Http post request
        response = requests.post(
            url=f"{self.__oracle_url}/faucet",
            json={"address": self.__acc_address},
            headers=self.__rest_header,
            timeout=20,  # 10
        )

        # raise Exception if status is not successful
        response.raise_for_status()

        return print("ORACLE: Received 500 ETH", flush=True)

    def verify_balance(self) -> None:
        """
        Calls blockchain directly for requesting current balance
        Returns: None

        """

        # directly call view method from non-validator node
        balance = self.__web3.eth.get_balance(self.__acc_address, "latest")

        # convert wei to ether
        balance_eth = self.__web3.from_wei(balance, "ether")
        print(
            f"BLOCKCHAIN: Successfully verified balance of {balance_eth} ETH",
            flush=True,
        )

        # if node ran out of funds, it requests ether from the oracle
        if balance_eth <= 1:
            self.__request_funds_from_oracle()

        return None

    @retry((Exception, requests.exceptions.HTTPError), tries=3, delay=4)
    def __get_contract_from_oracle(self):
        """
        Requests header file and contract address, generates Web3 Contract object with it
        Returns: Web3 Contract object

        """

        response = requests.get(
            url=f"{self.__oracle_url}/contract",
            headers=self.__rest_header,
            timeout=20,  # 10
        )

        # raise Exception if status is not successful
        response.raise_for_status()

        # convert response to json to extract the abi and address
        json_response = response.json()

        print(
            f"ORACLE: Initialized chain code: {json_response.get('address')}",
            flush=True,
        )

        # return an initialized web3 contract object
        return self.__web3.eth.contract(abi=json_response.get("abi"), address=json_response.get("address"))

    @retry((Exception, requests.exceptions.HTTPError), tries=3, delay=4)
    def report_gas_oracle(self) -> list:
        """
        Reports accumulated gas costs of all transactions made to the blockchain
        Returns: List of all accumulated gas costs per registered node

        """

        # method used for experiments, not needed for aggregation
        response = requests.post(
            url=f"{self.__oracle_url}/gas",
            json={"amount": self.__gas_used, "round": self.round},
            headers=self.__rest_header,
            timeout=20,  # 10
        )

        # raise Exception if status is not successful
        response.raise_for_status()

        # reset local gas accumulation
        self.__gas_used = 0

        # return list with gas usage for logging
        return list(response.json().items())

    @retry((Exception, requests.exceptions.HTTPError), tries=3, delay=4)
    def report_reputation_oracle(self, records: list) -> None:
        """
        Reports reputations used for aggregation
        Returns: None

        """

        # method used for experiments, not needed for aggregation
        response = requests.post(
            url=f"{self.__oracle_url}/reputation",
            json={"records": records, "round": self.round, "sender": self.__home_ip},
            headers=self.__rest_header,
            timeout=20,  # 10
        )

        # raise Exception if status is not successful
        response.raise_for_status()

        return None

    def __sign_and_deploy(self, trx_hash):
        """
        Signs a function call to the chain code with the primary key and awaits the receipt
        Args:
            trx_hash: Transformed dictionary of all properties relevant for call to chain code

        Returns: transaction receipt confirming the successful write to the ledger

        """

        # transaction is signed with private key
        signed_transaction = self.__web3.eth.account.sign_transaction(trx_hash, private_key=self.__private_key)

        # confirmation that transaction was passed from non-validator node to validator nodes
        executed_transaction = self.__web3.eth.send_raw_transaction(signed_transaction.rawTransaction)

        # non-validator node awaited the successful validation by validation nodes and returns receipt
        transaction_receipt = self.__web3.eth.wait_for_transaction_receipt(executed_transaction)

        # accumulate used gas
        self.__gas_used += transaction_receipt.gasUsed

        return transaction_receipt

    @retry(Exception, tries=3, delay=4)
    def push_opinions(self, opinion_dict: dict):
        """
        Pushes all locally computed opinions of models to aggregate to the reputation system
        Args:
            opinion_dict: Dict of all names:opinions for writing to the reputation system

        Returns: Json of transaction receipt

        """

        # create raw transaction object to call rate_neighbors() from the reputation system
        unsigned_trx = self.__contract_obj.functions.rate_neighbours(list(opinion_dict.items())).build_transaction({
            "chainId": self.__web3.eth.chain_id,
            "from": self.__acc_address,
            "nonce": self.__web3.eth.get_transaction_count(
                self.__web3.to_checksum_address(self.__acc_address), "pending"
            ),
            "gasPrice": self.__web3.to_wei(self.__gas_price, "gwei"),
        })

        # sign and execute the transaction
        conf = self.__sign_and_deploy(unsigned_trx)

        self.report_reputation_oracle(list(opinion_dict.items()))
        # return the receipt as json
        return self.__web3.to_json(conf)

    @retry(Exception, tries=3, delay=4)
    def get_reputations(self, ip_addresses: list) -> dict:
        """
        Requests globally aggregated opinions values from reputation system for computing aggregation weights
        Args:
            ip_addresses: Names of nodes of which the reputation values should be generated

        Returns: Dictionary of name:reputation from the reputation system

        """

        final_reputations = dict()
        stats_to_print = list()

        # call get_reputations() from reputation system
        raw_reputation = self.__contract_obj.functions.get_reputations(ip_addresses).call({"from": self.__acc_address})

        # loop list with tuples from reputation system
        for (
            name,
            reputation,
            weighted_reputation,
            stddev_count,
            divisor,
            final_reputation,
            avg,
            median,
            stddev,
            index,
            avg_deviation,
            avg_avg_deviation,
            malicious_opinions,
        ) in raw_reputation:
            # list elements with an empty name can be ignored
            if not name:
                continue

            # print statistical values
            stats_to_print.append([
                name,
                reputation / 10,
                weighted_reputation / 10,
                stddev_count / 10,
                divisor / 10,
                final_reputation / 10,
                avg / 10,
                median / 10,
                stddev / 10,
                avg_deviation / 10,
                avg_avg_deviation / 10,
                malicious_opinions,
            ])

            # assign the final reputation to a dict for later aggregation
            final_reputations[name] = final_reputation / 10

        print_table(
            "REPUTATION SYSTEM STATE",
            stats_to_print,
            [
                "Name",
                "Reputation",
                "Weighted Rep. by local Node",
                "Stddev Count",
                "Divisor",
                "Final Reputation",
                "Mean",
                "Median",
                "Stddev",
                "Avg Deviation in Opinion",
                "Avg of all Avg Deviations in Opinions",
                "Malicious Opinions",
            ],
        )

        # if sum(final_reputations.values()):
        #     self.report_reputation_oracle(list(final_reputations.items()))

        return final_reputations

    @retry(Exception, tries=3, delay=4)
    def __register(self) -> str:
        """
        Registers a node's name with its public address, signed with private key
        Returns: Json of transaction receipt

        """

        # build raw transaction object to call public method register() from reputation system
        unsigned_trx = self.__contract_obj.functions.register(self.__home_ip).build_transaction({
            "chainId": self.__web3.eth.chain_id,
            "from": self.__acc_address,
            "nonce": self.__web3.eth.get_transaction_count(
                self.__web3.to_checksum_address(self.__acc_address), "pending"
            ),
            "gasPrice": self.__web3.to_wei(self.__gas_price, "gwei"),
        })

        # sign and execute created transaction
        conf = self.__sign_and_deploy(unsigned_trx)

        # return the receipt as json
        return self.__web3.to_json(conf)

    @retry(Exception, tries=3, delay=4)
    def verify_registration(self) -> None:
        """
        Verifies the successful registration of the node itself,
        executes registration again if reputation system returns false
        Returns: None

        """

        # call view function of reputation system to check if registration was not abandoned by hard fork
        confirmation = self.__contract_obj.functions.confirm_registration().call({"from": self.__acc_address})

        # function returns boolean
        if not confirmation:
            # register again if not successful
            self.__register()

            # raise Exception to check again
            raise Exception("EXCEPTION: _verify_registration() => Could not be confirmed)")

        return None

    @retry((Exception, requests.exceptions.HTTPError), tries=3, delay=4)
    def report_time_oracle(self, start: float) -> None:
        """
        Reports time used for aggregation
        Returns: None

        """
        # method used for experiments, not needed for aggregation
        # report aggregation time and round to oracle
        response = requests.post(
            url=f"{BlockchainHandler.oracle_url}/time",
            json={"time": (time.time_ns() - start) / (10**9), "round": self.round},
            headers=self.__rest_header,
            timeout=20,  # 10
        )

        # raise Exception if status is not successful
        response.raise_for_status()

        # increase aggregation round counter after reporting time
        self.round += 1
        return None



import asyncio
import bz2
import json
import logging
import lzma
import time
import uuid
import zlib
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import lz4.frame
from geopy import distance

if TYPE_CHECKING:
    from nebula.core.network.communications import CommunicationsManager


@dataclass
class MessageChunk:
    __slots__ = ["index", "data", "is_last"]
    index: int
    data: bytes
    is_last: bool


class Connection:
    DEFAULT_FEDERATED_ROUND = -1

    def __init__(
        self,
        cm: "CommunicationsManager",
        reader,
        writer,
        id,
        host,
        port,
        direct=True,
        active=True,
        compression="zlib",
        config=None,
    ):
        self.cm = cm
        self.reader = reader
        self.writer = writer
        self.id = str(id)
        self.host = host
        self.port = port
        self.addr = f"{host}:{port}"
        self.direct = direct
        self.active = active
        self.last_active = time.time()
        self.compression = compression
        self.config = config

        self.federated_round = Connection.DEFAULT_FEDERATED_ROUND
        self.latitude = None
        self.longitude = None
        self.loop = asyncio.get_event_loop()
        self.read_task = None
        self.process_task = None
        self.pending_messages_queue = asyncio.Queue(maxsize=100)
        self.message_buffers: dict[bytes, dict[int, MessageChunk]] = {}

        self.EOT_CHAR = b"\x00\x00\x00\x04"
        self.COMPRESSION_CHAR = b"\x00\x00\x00\x01"
        self.DATA_TYPE_PREFIXES = {
            "pb": b"\x01\x00\x00\x00",
            "string": b"\x02\x00\x00\x00",
            "json": b"\x03\x00\x00\x00",
            "bytes": b"\x04\x00\x00\x00",
        }
        self.HEADER_SIZE = 21
        self.MAX_CHUNK_SIZE = 1024  # 1 KB
        self.BUFFER_SIZE = 1024  # 1 KB

        logging.info(
            f"Connection [established]: {self.addr} (id: {self.id}) (active: {self.active}) (direct: {self.direct})"
        )

    def __str__(self):
        return f"Connection to {self.addr} (id: {self.id}) (active: {self.active}) (last active: {self.last_active}) (direct: {self.direct})"

    def __repr__(self):
        return self.__str__()

    def __del__(self):
        self.stop()

    def get_addr(self):
        return self.addr

    def get_federated_round(self):
        return self.federated_round

    def get_tunnel_status(self):
        if self.reader is None or self.writer is None:
            return False
        return True

    def update_round(self, federated_round):
        self.federated_round = federated_round

    def update_geolocation(self, latitude, longitude):
        self.latitude = latitude
        self.longitude = longitude
        self.config.participant["mobility_args"]["neighbors_distance"][self.addr] = self.compute_distance_myself()

    def get_geolocation(self):
        if self.latitude is None or self.longitude is None:
            raise ValueError("Geo-location not set for this neighbor")
        return self.latitude, self.longitude

    def get_neighbor_distance(self):
        if self.addr not in self.config.participant["mobility_args"]["neighbors_distance"]:
            return None
        return self.config.participant["mobility_args"]["neighbors_distance"][self.addr]

    def compute_distance(self, latitude, longitude):
        distance_m = distance.distance((self.latitude, self.longitude), (latitude, longitude)).m
        return distance_m

    def compute_distance_myself(self):
        distance_m = self.compute_distance(
            self.config.participant["mobility_args"]["latitude"],
            self.config.participant["mobility_args"]["longitude"],
        )
        return distance_m

    def get_ready(self):
        return True if self.federated_round != Connection.DEFAULT_FEDERATED_ROUND else False

    def get_direct(self):
        return self.direct

    def set_direct(self, direct):
        # config.participant["network_args"]["neighbors"] only contains direct neighbors (frotend purposes)
        if direct:
            self.config.add_neighbor_from_config(self.addr)
        else:
            self.config.remove_neighbor_from_config(self.addr)
        self.last_active = time.time()
        self.direct = direct

    def set_active(self, active):
        self.active = active
        self.last_active = time.time()

    def is_active(self):
        return self.active

    def get_last_active(self):
        return self.last_active

    async def start(self):
        self.read_task = asyncio.create_task(self.handle_incoming_message(), name=f"Connection {self.addr} reader")
        self.process_task = asyncio.create_task(self.process_message_queue(), name=f"Connection {self.addr} processor")

    async def stop(self):
        logging.info(f"❗️  Connection [stopped]: {self.addr} (id: {self.id})")
        tasks = [self.read_task, self.process_task]
        for task in tasks:
            if task is not None:
                task.cancel()
                try:
                    await task
                except asyncio.CancelledError:
                    logging.exception(f"❗️  {self} cancelled...")

        if self.writer is not None:
            self.writer.close()
            await self.writer.wait_closed()

    async def reconnect(self, max_retries: int = 5, delay: int = 5) -> None:
        for attempt in range(max_retries):
            try:
                logging.info(f"Attempting to reconnect to {self.addr} (attempt {attempt + 1}/{max_retries})")
                await self.cm.connect(self.addr)
                self.read_task = asyncio.create_task(
                    self.handle_incoming_message(),
                    name=f"Connection {self.addr} reader",
                )
                self.process_task = asyncio.create_task(
                    self.process_message_queue(),
                    name=f"Connection {self.addr} processor",
                )
                logging.info(f"Reconnected to {self.addr}")
                return
            except Exception as e:
                logging.exception(f"Reconnection attempt {attempt + 1} failed: {e}")
                await asyncio.sleep(delay)
        logging.error(f"Failed to reconnect to {self.addr} after {max_retries} attempts. Stopping connection...")
        await self.stop()

    async def send(
        self,
        data: Any,
        pb: bool = True,
        encoding_type: str = "utf-8",
        is_compressed: bool = False,
    ) -> None:
        if self.writer is None:
            logging.error("Cannot send data, writer is None")
            return

        try:
            message_id = uuid.uuid4().bytes
            data_prefix, encoded_data = self._prepare_data(data, pb, encoding_type)

            if is_compressed:
                encoded_data = await asyncio.to_thread(self._compress, encoded_data, self.compression)
                if encoded_data is None:
                    return
                data_to_send = data_prefix + encoded_data + self.COMPRESSION_CHAR
            else:
                data_to_send = data_prefix + encoded_data

            await self._send_chunks(message_id, data_to_send)
        except Exception as e:
            logging.exception(f"Error sending data: {e}")
            await self.reconnect()

    def _prepare_data(self, data: Any, pb: bool, encoding_type: str) -> tuple[bytes, bytes]:
        if pb:
            return self.DATA_TYPE_PREFIXES["pb"], data
        elif isinstance(data, str):
            return self.DATA_TYPE_PREFIXES["string"], data.encode(encoding_type)
        elif isinstance(data, dict):
            return self.DATA_TYPE_PREFIXES["json"], json.dumps(data).encode(encoding_type)
        elif isinstance(data, bytes):
            return self.DATA_TYPE_PREFIXES["bytes"], data
        else:
            raise ValueError(f"Unknown data type to send: {type(data)}")

    def _compress(self, data: bytes, compression: str) -> bytes | None:
        if compression == "lz4":
            return lz4.frame.compress(data)
        elif compression == "zlib":
            return zlib.compress(data)
        elif compression == "bz2":
            return bz2.compress(data)
        elif compression == "lzma":
            return lzma.compress(data)
        else:
            logging.error(f"Unsupported compression method: {compression}")
            return None

    async def _send_chunks(self, message_id: bytes, data: bytes) -> None:
        chunk_size = self._calculate_chunk_size(len(data))
        num_chunks = (len(data) + chunk_size - 1) // chunk_size

        for chunk_index in range(num_chunks):
            start = chunk_index * chunk_size
            end = min(start + chunk_size, len(data))
            chunk = data[start:end]
            is_last_chunk = chunk_index == num_chunks - 1

            header = message_id + chunk_index.to_bytes(4, "big") + (b"\x01" if is_last_chunk else b"\x00")
            chunk_size_bytes = len(chunk).to_bytes(4, "big")
            chunk_with_header = header + chunk_size_bytes + chunk + self.EOT_CHAR

            self.writer.write(chunk_with_header)
            await self.writer.drain()

            # logging.debug(f"Sent message {message_id.hex()} | chunk {chunk_index+1}/{num_chunks} | size: {len(chunk)} bytes")

    def _calculate_chunk_size(self, data_size: int) -> int:
        return self.BUFFER_SIZE

    async def handle_incoming_message(self) -> None:
        reusable_buffer = bytearray(self.MAX_CHUNK_SIZE)
        try:
            while True:
                if self.pending_messages_queue.full():
                    await asyncio.sleep(0.1)  # Wait a bit if the queue is full to create backpressure
                    continue
                header = await self._read_exactly(self.HEADER_SIZE)
                message_id, chunk_index, is_last_chunk = self._parse_header(header)

                chunk_data = await self._read_chunk(reusable_buffer)
                self._store_chunk(message_id, chunk_index, chunk_data, is_last_chunk)
                # logging.debug(f"Received chunk {chunk_index} of message {message_id.hex()} | size: {len(chunk_data)} bytes")

                if is_last_chunk:
                    await self._process_complete_message(message_id)
        except asyncio.CancelledError:
            logging.info("Message handling cancelled")
        except ConnectionError as e:
            logging.exception(f"Connection closed while reading: {e}")
            await self.reconnect()
        except Exception as e:
            logging.exception(f"Error handling incoming message: {e}")

    async def _read_exactly(self, num_bytes: int, max_retries: int = 3) -> bytes:
        data = b""
        remaining = num_bytes
        for _ in range(max_retries):
            try:
                while remaining > 0:
                    chunk = await self.reader.read(min(remaining, self.BUFFER_SIZE))
                    if not chunk:
                        raise ConnectionError("Connection closed while reading")
                    data += chunk
                    remaining -= len(chunk)
                return data
            except asyncio.IncompleteReadError as e:
                if _ == max_retries - 1:
                    raise
                logging.warning(f"Retrying read after IncompleteReadError: {e}")
        raise RuntimeError("Max retries reached in _read_exactly")

    def _parse_header(self, header: bytes) -> tuple[bytes, int, bool]:
        message_id = header[:16]
        chunk_index = int.from_bytes(header[16:20], "big")
        is_last_chunk = header[20] == 1
        return message_id, chunk_index, is_last_chunk

    async def _read_chunk(self, buffer: bytearray = None) -> bytes:
        if buffer is None:
            buffer = bytearray(self.MAX_CHUNK_SIZE)

        chunk_size_bytes = await self._read_exactly(4)
        chunk_size = int.from_bytes(chunk_size_bytes, "big")

        if chunk_size > self.MAX_CHUNK_SIZE:
            raise ValueError(f"Chunk size {chunk_size} exceeds MAX_CHUNK_SIZE {self.MAX_CHUNK_SIZE}")

        chunk = await self._read_exactly(chunk_size)
        buffer[:chunk_size] = chunk
        eot = await self._read_exactly(len(self.EOT_CHAR))

        if eot != self.EOT_CHAR:
            raise ValueError("Invalid EOT character")

        return memoryview(buffer)[:chunk_size]

    def _store_chunk(self, message_id: bytes, chunk_index: int, buffer: memoryview, is_last: bool) -> None:
        if message_id not in self.message_buffers:
            self.message_buffers[message_id] = {}
        try:
            self.message_buffers[message_id][chunk_index] = MessageChunk(chunk_index, buffer.tobytes(), is_last)
            # logging.debug(f"Stored chunk {chunk_index} of message {message_id.hex()} | size: {len(data)} bytes")
        except Exception as e:
            if message_id in self.message_buffers:
                del self.message_buffers[message_id]
            logging.exception(f"Error storing chunk {chunk_index} for message {message_id.hex()}: {e}")

    async def _process_complete_message(self, message_id: bytes) -> None:
        chunks = sorted(self.message_buffers[message_id].values(), key=lambda x: x.index)
        complete_message = b"".join(chunk.data for chunk in chunks)
        del self.message_buffers[message_id]

        data_type_prefix = complete_message[:4]
        message_content = complete_message[4:]

        if message_content.endswith(self.COMPRESSION_CHAR):
            message_content = await asyncio.to_thread(
                self._decompress,
                message_content[: -len(self.COMPRESSION_CHAR)],
                self.compression,
            )
            if message_content is None:
                return

        await self.pending_messages_queue.put((data_type_prefix, memoryview(message_content)))
        # logging.debug(f"Processed complete message {message_id.hex()} | total size: {len(complete_message)} bytes")

    def _decompress(self, data: bytes, compression: str) -> bytes | None:
        if compression == "zlib":
            return zlib.decompress(data)
        elif compression == "bz2":
            return bz2.decompress(data)
        elif compression == "lzma":
            return lzma.decompress(data)
        elif compression == "lz4":
            return lz4.frame.decompress(data)
        else:
            logging.error(f"Unsupported compression method: {compression}")
            return None

    async def process_message_queue(self) -> None:
        while True:
            try:
                if self.pending_messages_queue is None:
                    logging.error("Pending messages queue is not initialized")
                    return
                data_type_prefix, message = await self.pending_messages_queue.get()
                await self._handle_message(data_type_prefix, message)
                self.pending_messages_queue.task_done()
            except Exception as e:
                logging.exception(f"Error processing message queue: {e}")
            finally:
                await asyncio.sleep(0)

    async def _handle_message(self, data_type_prefix: bytes, message: bytes) -> None:
        if data_type_prefix == self.DATA_TYPE_PREFIXES["pb"]:
            # logging.debug("Received a protobuf message")
            asyncio.create_task(
                self.cm.handle_incoming_message(message, self.addr),
                name=f"Connection {self.addr} message handler",
            )
        elif data_type_prefix == self.DATA_TYPE_PREFIXES["string"]:
            logging.debug(f"Received string message: {message.decode('utf-8')}")
        elif data_type_prefix == self.DATA_TYPE_PREFIXES["json"]:
            logging.debug(f"Received JSON message: {json.loads(message.decode('utf-8'))}")
        elif data_type_prefix == self.DATA_TYPE_PREFIXES["bytes"]:
            logging.debug(f"Received bytes message of length: {len(message)}")
        else:
            logging.error(f"Unknown data type prefix: {data_type_prefix}")



import asyncio
import logging
from typing import TYPE_CHECKING

from nebula.addons.functions import print_msg_box
from nebula.core.pb import nebula_pb2

if TYPE_CHECKING:
    from nebula.core.network.communications import CommunicationsManager


class Discoverer:
    def __init__(self, addr, config, cm: "CommunicationsManager"):
        print_msg_box(msg="Starting discoverer module...", indent=2, title="Discoverer module")
        self.addr = addr
        self.config = config
        self.cm = cm
        self.grace_time = self.config.participant["discoverer_args"]["grace_time_discovery"]
        self.period = self.config.participant["discoverer_args"]["discovery_frequency"]
        self.interval = self.config.participant["discoverer_args"]["discovery_interval"]

    async def start(self):
        asyncio.create_task(self.run_discover())

    async def run_discover(self):
        if self.config.participant["scenario_args"]["federation"] == "CFL":
            logging.info("🔍  Federation is CFL. Discoverer is disabled...")
            return
        await asyncio.sleep(self.grace_time)
        while True:
            if len(self.cm.connections) > 0:
                latitude = self.config.participant["mobility_args"]["latitude"]
                longitude = self.config.participant["mobility_args"]["longitude"]
                message = self.cm.mm.generate_discovery_message(
                    action=nebula_pb2.DiscoveryMessage.Action.DISCOVER,
                    latitude=latitude,
                    longitude=longitude,
                )
                try:
                    logging.debug("🔍  Sending discovery message to neighbors...")
                    current_connections = await self.cm.get_addrs_current_connections(only_direct=True)
                    await self.cm.send_message_to_neighbors(message, current_connections, self.interval)
                except Exception as e:
                    logging.exception(f"🔍  Cannot send discovery message to neighbors. Error: {e!s}")
            await asyncio.sleep(self.period)
